///////////////////////////////////////////////////////////////////////////////
// $Revision: 1.136.2.5.4.1 $
///////////////////////////////////////////////////////////////////////////////
#pragma option(strict,on)

/*
  Don't modify this code unless defining extension specific
  aliases does not suite your needs.   For example, if you
  want your brace style to be:

       if () {
          }

  Use the Extension Options dialog box ("Other", "Configuration...",
  "File Extension Setup...") and press the the "Alias" button to
  display the Alias Editor dialog box.  Press the New button, type
  "if" for the name of the alias and press <Enter>.  Enter the
  following text into the upper right editor control:

       if (%\c) {
       %\i}

  The  %\c indicates where the cursor will be placed after the
  "if" alias expanded.  The %\i specifies to indent by the
  Extension Specific "Syntax Indent" amount define in the
  Extension Options dialog box.  Check the "Indent With Tabs"
  check box on the Extension Options dialog box if you want
  the %\i option to indent using tab characters.

*/
/*
  Options for C syntax expansion/indenting may be accessed from the
  Extension Options dialog ("Other", "Configuration...",
  "File Extension Setup...").

  The extension specific options is a string of five numbers separated
  with spaces with the following meaning:

    Position       Option
       1             Syntax indent amount
       2             expansion on/off.
       3             Minimum abbreviation.  Defaults to 1.  Specify large
                     value to avoid abbreviation expansion.
       4             Indent after open parenthesis.  Effects argument
                     lists and if/while/switch.
       5             begin/end style.  Begin/end style may be 0,1, or 2
                     as show below.  Add 4 to the begin end style if you
                     want braces inserted when syntax expansion occurs
                     (main and do insert braces anyway).  Typing a begin
                     brace, '{', inserts an end brace when appropriate
                     (unless you unbind the key).  If you want a blank
                     line inserted in between, add 8 to the begin end
                     style.  Default is 4.

                      Style 0
                          if () {
                             ++i;
                          }

                      Style 1
                          if ()
                          {
                             ++i;
                          }

                      Style 2
                          if ()
                            {
                            ++i;
                            }


       6             Indent first level of code.  Default is 1.
                     Specify 0 if you want first level statements to
                     start in column 1.
       7             Main style.  Main style may be 0,1, or 2 which
                     correspond to old C style, ansi C, or no expansion.
                     Default is 0.
       8             Indent CASE from SWITCH.  Default is 0.  Specify
                     1 if you want CASE statements indented from the
                     SWITCH statement. Begin/end style 2 not supported.
       9             Not always present.
                     UseContOnParameters
*/

#include 'slick.sh'
#include 'tagsdb.sh'

#define TK_ID 1
#define TK_NUMBER 2
#define TK_STRING 3
#define TK_KEYWORD 4


// style3 refers to out C extension options dialog
int def_style3_indent_all_braces;

// always expand braces in c_begin
boolean def_expand_brace_syntax=true;

// always prompt for else or else if,
// even if "else" is completely typed out
boolean def_always_prompt_for_else_if=false;


/**
 * Activates C/C++ file editing mode.
 *
 * @appliesTo	Edit_Window, Editor_Control
 *
 * @categories Edit_Window_Methods, Editor_Control_Methods
 */
_command void c_mode() name_info(','VSARG2_REQUIRES_EDITORCTL|VSARG2_READ_ONLY|VSARG2_ICON)
{
   select_edit_mode('c');
}

/**
 * (C mode only) ENTER
 * <pre>
 * New binding of ENTER key when in C mode.  Handles syntax expansion and indenting for files
 * with C or H extension.  Use Extension Options dialog box ("Tools", "Options", "File Extension Setup...",
 * select the Indent tab) to change syntax expansion/indenting options.
 *
 * @appliesTo	Edit_Window, Editor_Control
 *
 * @categories Edit_Window_Methods, Editor_Control_Methods
 */
_command void c_enter() name_info(','VSARG2_CMDLINE|VSARG2_ICON|VSARG2_REQUIRES_EDITORCTL|VSARG2_READ_ONLY)
{  
   if (ispf_common_enter()) return;
   typeless expand, be_style, indent_fl, indent_case;
   parse name_info(_edit_window().p_index) with . expand . . be_style indent_fl . indent_case .;
   if (command_state()) {
      call_root_key(ENTER);
      return;
   }

   // Handle Assembler embedded in C
   typeless orig_values;
   int embedded_status=_EmbeddedStart(orig_values,'');
   if (embedded_status==1) {
      call_key(ENTER);
      _EmbeddedEnd(orig_values);
      return; // Processing done for this key
   }
   if (p_window_state:=='I' ||
      p_SyntaxIndent<0 || p_indent_style!=INDENT_SMART) {
      call_root_key(ENTER);
   } else {
      if (_in_comment(true)) {
         // start of a Java doc comment?
         _str first_line;
         get_line(first_line);
         if (_ext_comment_editing_flags(VS_COMMENT_EDITING_FLAG_AUTO_JAVADOC_COMMENT) &&
             first_line=='/***/' && get_text(2)=='*/' && _is_line_before_decl()) {
            javadoc_comment();
            commentwrap_SetNewJavadocState();
            return;
         }
         //Try to handle with comment wrap.  If coment wrap
         //handled the keystroke then return.
         if (commentwrap_Enter()) {
            return;
         }
         // multi-line comment
         int star_col = _inJavadoc();
         if (_ext_comment_editing_flags(VS_COMMENT_EDITING_FLAG_AUTO_JAVADOC_ASTERISK) && star_col>0) {
            indent_on_enter(0,star_col);
            if (get_text()!='*') {
               keyin('* ');
            }
         } else {
            call_root_key(ENTER);
         }
         return;
      }
      if (_in_comment(false)) {
         // single line comment
         
         //Try to handle with comment wrap.  If coment wrap
         //handled the keystroke then return.
         if (commentwrap_Enter()) {
            return;
         }
         _str commentChars='';
         int line_col = _inExtendableLineComment(commentChars);
         if (_ext_comment_editing_flags(VS_COMMENT_EDITING_FLAG_SPLIT_LINE_COMMENTS) &&
             line_col && _will_split_insert_line() && get_text()!='/') {
            // check for xmldoc comment
            int orig_col=p_col;
            p_col = line_col;
            boolean triple_slash = (get_text(3)=='///' && get_text(4)!='////');
            p_col = orig_col;
            if (_ext_comment_editing_flags(VS_COMMENT_EDITING_FLAG_AUTO_XMLDOC_COMMENT) &&
                triple_slash && (ext_inherits_from('cs') || ext_inherits_from('c') || ext_inherits_from('jsl')) &&
                c_maybe_create_xmldoc_comment()) {
               return;
            }
            indent_on_enter(0,line_col);
            if (get_text(2)!='//') {
               if (triple_slash) {
                  keyin('/// ');
               } else {
                  keyin('// ');
               }
            }
            return;
         }
      }
      if (_in_string()) {
         if (ext_inherits_from('c') || ext_inherits_from('cs') || ext_inherits_from('java')) {
            _str delim='';
            int string_col = _inString(delim);
            if (_ext_comment_editing_flags(VS_COMMENT_EDITING_FLAG_SPLIT_STRINGS) && 
                string_col && p_col > string_col && _will_split_insert_line()) {
               _insert_text(delim);
               if (ext_inherits_from('java')) _insert_text('+');
               if (ext_inherits_from('cs')) _insert_text('+');
               indent_on_enter(0,string_col);
               keyin(delim);
               return;
            }
         }
      }
      if (_c_expand_enter(p_SyntaxIndent,expand,be_style,indent_fl,indent_case) ) {
          call_root_key(ENTER);
      } else if (_argument=='') {
         _undo('S');
      }
   }
}
static int find_class_col()
{
   typeless p;
   save_pos(p);
   int status=search('class|[{}]','@-r');
   int nest_level=0;
   for (;;) {
      if (status) {
         restore_pos(p);
         return(0);
      }
      int cfg=_clex_find(0,'g');
      if (nest_level>=0 || cfg!=CFG_KEYWORD) {
         _str ch=get_text();
         if (ch=='{') {
            --nest_level;
         } else if (ch=='}') {
            ++nest_level;
         }
         status=repeat_search();
         continue;
      }
      first_non_blank();
      int col=p_col;
      restore_pos(p);
      return(col);
   }
}
boolean _c_do_colon()
{
   if (p_col<=2) return(true);
   int orig_col=p_col;
   left();left();
   _str word, junk, line;
   int cfg=_clex_find(0,'g');
   get_line(line);
   _str maybe_slots,rest;
   parse line with word maybe_slots rest;
   if (cfg==CFG_KEYWORD || (line=='signals:' && ext_inherits_from('c')) || 
       ( maybe_slots=='slots:' && ext_inherits_from('c') &&
         (word=='public' || word=='private' || word=='protected') && rest==''
        ) 
      ) {
      word=cur_word(junk);
      if (word=='public' || word=='private' || word=='protected' ||
          word=='slots' || word=='signals'
          ) {
         if (word!='slots') {
            first_non_blank();
            if (p_col!=orig_col-length(word)-1) {
               p_col=orig_col;
               return(true);
            }
         }
         int class_col=find_class_col();
         if (!class_col) {
            p_col=orig_col;
            return(true);
         }
         get_line(line);line=strip(line);
         if (def_style3_indent_all_braces) {
            class_col+=p_SyntaxIndent;
         }
         replace_line(indent_string(class_col-1):+strip(line,'L'));
         if (word=='slots') {
            _end_line();
         } else {
            p_col=class_col+length(word)+1;
         }

         return(false);
      }
   }

   int orig_linenum=p_line;
   typeless p;
   save_pos(p);
   int begin_col=c_begin_stat_col(false /* No RestorePos */,
                                  false /* Don't skip first begin statement marker */,
                                  false /* Don't return first non-blank */);
   // IF we found the beginning of this statement and it starts on
   //    the same line as the colon
   if (begin_col && p_line==orig_linenum) {
      word=cur_word(junk);
      if (word=='case' || word=='default') {
         first_non_blank();
         // IF the 'case' word is the first non-blank on this line
         if (p_col==begin_col) {
            typeless expand, be_style, indent_fl, indent_case;
            parse name_info(_edit_window().p_index) with . expand . . be_style indent_fl . indent_case .;
            _str cur_line;
            get_line_raw(cur_line);
            int col=_c_last_switch_col();
            if ( col) {
               if ((indent_case && indent_case!='') || (be_style&VS_C_OPTIONS_STYLE2_FLAG)) {
                  col=col+p_SyntaxIndent;
               }
               _str new_cur_line=indent_string(col-1):+strip(cur_line,'L');
               replace_line_raw(new_cur_line);
               // adjust cursor column based on new length of line
               p_col=orig_col+length(expand_tabs(new_cur_line))-length(expand_tabs(cur_line));
            } else {
               p_col=orig_col;
            }
            return(false);
         }
      }
   }
   restore_pos(p);

   p_col=orig_col;
   return(true);
}
_command void c_colon() name_info(','VSARG2_CMDLINE|VSARG2_REQUIRES_EDITORCTL)
{
   keyin(':');
   int cfg=0;
   typeless expand, be_style, indent_fl;
   parse name_info(_edit_window().p_index) with . expand . . be_style indent_fl . ;
   if (!command_state()) {
      left();
      cfg=_clex_find(0,'g');
      right();
   }
   if ( command_state() || p_SyntaxIndent<0 ||
      _in_comment() || cfg==CFG_STRING) {
   } else {
      if (_c_do_colon() &&
          (_ext_codehelp_flags() & VSCODEHELPFLAG_AUTO_LIST_MEMBERS)
         ) {
         _do_list_members(true,false);
      }
   /*} else if (_argument=='') {
      _undo('S');*/
   }
}
_command void c_pound() name_info(','VSARG2_CMDLINE|VSARG2_REQUIRES_EDITORCTL|VSARG2_LASTKEY)
{
   if (command_state()) {
      call_root_key(last_event());
      return;
   }
   if ( !(_ext_codehelp_flags() & VSCODEHELPFLAG_AUTO_LIST_MEMBERS)) {
      keyin('#');
      return;
   }
   if (_in_comment() && _inJavadocSeeTag()) {
      auto_codehelp_key();
      return;
   }
   keyin('#');
#if 0
   if (_in_comment()) {
      return;
   }
   parse name_info(_edit_window().p_index) with . expand . . be_style indent_fl . ;
   c_expand_space(p_SyntaxIndent,be_style,indent_fl);
#endif
}
/**
 * If the user just typed <code>/*</code> to open a multiple line comment,
 * complete the comment with <code>*/</code>.  Do not complete the comment
 * if there is text after the cursor, since they are probably trying to
 * surround something.
 * 
 * @appliesTo	Edit_Window, Editor_Control
 * 
 * @categories Edit_Window_Methods, Editor_Control_Methods
 */
_command void c_asterisk() name_info(','VSARG2_CMDLINE|VSARG2_REQUIRES_EDITORCTL)
{
   if (command_state()) {
      call_root_key(last_event());
      return;
   }
   if (_in_comment() || _in_string()) {
      boolean doInsert = false;
      if (p_col>2) {
         p_col-=2;_str temp = get_text(4);p_col+=2;
         if (temp == '/**/') {
            doInsert = true;
         }
      }
      if (doInsert) {
         _insert_text('*');
      } else {
         keyin('*');
      }
      return;
   }

   _str orig_line='';
   get_line(orig_line);
   _str line=strip(orig_line,'T');
   if ( p_col != text_col(_rawText(line))+1 ) {
      keyin('*');
      return;
   }

   typeless p;
   save_pos(p);
   left();
   if (get_text() == '/' && search('*/','@Xcs') < 0) {
      restore_pos(p);
		_insert_text_raw('**/');
		p_col -= 2;
		message("Type '*/' on a subsequent line to finish this block comment.");
		return;
   }

   restore_pos(p);
   keyin('*');
}

/**
 * Is the location under the cursor a reasonable place for an
 * xmlDoc style comment?
 */
static boolean c_maybe_create_xmldoc_comment()
{
   // save our current position
   typeless p;
   save_pos(p);

   // break look for default handling
   do {

      // option is disabled
      if (!_ext_comment_editing_flags(VS_COMMENT_EDITING_FLAG_AUTO_XMLDOC_COMMENT)) break;

      // multiple-line comment or string?
      if (_in_comment(true) || _in_string()) break;

      // not in a single line comment?
      if (!_in_comment()) break;

      // not C#?
      if (!ext_inherits_from('cs') && 
          !ext_inherits_from('c')  &&
          !ext_inherits_from('jsl')) break;

      // the line must start with //
      _str line='';
      get_line(line);
      if ( substr(strip(line),1,2)!='//') break;
      
      // cursor must be at the end of line
      line=strip(line,'T');
      if ( p_col != text_col(_rawText(line))+1 ) break;

      // the previous line can not have a comment
      up(); _end_line();
      if (_in_comment()) break;

      // the next line can not have a comment
      restore_pos(p);
      down(); _end_line();
      if (_in_comment()) break;

      // update the current context
      restore_pos(p);
      if (!_is_line_before_decl()) break;

      // finally, we are ready to lay down a really cool comment
      xmldoc_comment();
      search("<summary>","-@Cc");
      down();
      return true;

   } while ( false );

   // handle like normal slash
   restore_pos(p);
   return false;
}

/**
 * If they just typed <code>///</code> to open a xmlDoc comment,
 * and we are not already inside a xmlDoc comment, and the current
 * scope is outside of a function, create a skeleton xmlDoc comment.
 * <p>
 * This feature is enabled for C# only
 * 
 * @see xmldoc_comment
 * 
 * @appliesTo	Edit_Window, Editor_Control
 * 
 * @categories Edit_Window_Methods, Editor_Control_Methods 
 */
_command void c_slash() name_info(','VSARG2_CMDLINE|VSARG2_REQUIRES_EDITORCTL)
{
   // cursor is on command line?
   if (command_state()) {
      call_root_key(last_event());
      return;
   }

   // if the previous character is '*' and we are not in
   // a string or comment, then the slash should look for
   // a /**/ that needs to be fixed to create a block comment.
   int cfg = _clex_find(0,'g');
   if (p_col > 1 && get_text(1,(int)_QROffset()-1)=='*' &&
       cfg!=CFG_STRING && cfg!=CFG_COMMENT && cfg!=CFG_IMAGINARY_LINE) {

      int start_line=0;
      int orig_line=p_line;
      typeless s1,s2,s3,s4,s5;
      save_search(s1,s2,s3,s4,s5);
      long orig_offset = _QROffset();

      int status = search('\/\*\*\/[ \t]*$','@-rCc');
      if (!status) {
         start_line=p_line;
         p_col+=2;
         _delete_text(2);
      }

      restore_search(s1,s2,s3,s4,s5);
      if (start_line > 0) {
         _GoToROffset(orig_offset-2);
         message("Finished comment starting on line ":+start_line);
      } else {
         _GoToROffset(orig_offset);
      }

      keyin('/');
      return;
   }

   // the line must only contain the leading // of the comment
   _str line='';
   get_line(line);
   if (strip(line)!='//' || !ext_inherits_from('cs')) {
      // handle like normal slash
      keyin('/');
      return;
   }

   // see if we can create an xmlDoc comment
   if (!c_maybe_create_xmldoc_comment()) {
      keyin('/');
      return;
   }
}

/**
 * (C mode only) SPACE BAR
 * <p>
 * New binding of SPACE key when in C mode.  Handles syntax expansion and indenting
 * for files with C or H extension.  Use the <b>C Options dialog box</b> ("Tools",
 * "Options", "File Extension Setup...", select C extension, "Options...")
 * to change the main style.
 *
 * @appliesTo	Edit_Window, Editor_Control
 *
 * @categories Edit_Window_Methods, Editor_Control_Methods
 *
 */
_command void c_space() name_info(','VSARG2_CMDLINE|VSARG2_REQUIRES_EDITORCTL|VSARG2_LASTKEY)
{
   // Handle Assembler embedded in C
   if (command_state()) {
      call_root_key(' ');
      return;
   }
   typeless orig_values;
   int embedded_status=_EmbeddedStart(orig_values,'');
   if (embedded_status==1) {
      call_key(last_event());
      _EmbeddedEnd(orig_values);
      return; // Processing done for this key
   }
   if (_inJavadoc()) {
      auto_codehelp_key();
      return;
   }
   typeless expand, be_style, indent_fl;
   parse name_info(_edit_window().p_index) with . expand . . be_style indent_fl .;
   if ( command_state() || ! expand || p_SyntaxIndent<0 ||
      _in_comment() ||
         c_expand_space(p_SyntaxIndent,be_style,indent_fl) ) {
      if ( command_state() ) {
         call_root_key(' ');
      } else {
         keyin(' ');
      }
   } else if (_argument=='') {
      _undo('S');
   }

   // display auto-list-paramters for completing
   // assignment statements, return statements,
   // and goto statements
   if (!command_state()) {
      c_maybe_list_args(true);
   }
}

/**
 * (C mode only) '{'
 * <p>
 * New binding of '{' key when in C mode.  When appropriate, inserts '{' at cursor,
 * inserts blank line, indents cursor on next line, and inserts another line with end brace.
 * Exact action depends on your C extension specific options.  These options may be
 * accessed by ("Tools", "Options", "File Extension Setup...", select C extension, "Options:").
 *
 *
 * @appliesTo	Edit_Window, Editor_Control
 *
 * @categories Edit_Window_Methods, Editor_Control_Methods
 */
_command void c_begin() name_info(','VSARG2_REQUIRES_EDITORCTL|VSARG2_CMDLINE)
{
   typeless expand, be_style, indent_fl;
   parse name_info(_edit_window().p_index) with . expand . . be_style indent_fl .;
   int cfg = 0;
   if (!command_state() && p_col>1) {
      left();cfg=_clex_find(0,'g');right();
   }
   if ( command_state() || cfg==CFG_STRING || _in_comment() ||
       c_expand_begin(expand,p_SyntaxIndent,be_style,indent_fl) ) {
      call_root_key('{');
   } else if (_argument=='') {
      _undo('S');
   }

}
/**
 * (C mode only) '}'
 * <pre>
 * New binding of '}' key when in C mode.  Inserts '}' at cursor and reindents the brace
 * according to your C extension specific options.  No reindenting occurs if the brace is
 * NOT being inserted into a blank line or if the brace is NOT part of an if, switch, while,
 * do, or else block.  The C extension specific options may be accessed by ("Tools",
 * "Options", "File Extension Setup...", select C extension, "Options:").
 *
 *
 * @appliesTo	Edit_Window, Editor_Control
 *
 * @categories Edit_Window_Methods, Editor_Control_Methods
 */
_command void c_endbrace() name_info(','VSARG2_CMDLINE|VSARG2_REQUIRES_EDITORCTL)
{
   typeless expand, be_style, indent_fl, indent_case;
   parse name_info(_edit_window().p_index) with . expand . . be_style indent_fl . indent_case . ;
   int cfg=0;
   if (!command_state() && p_col>1) {
      left();cfg=_clex_find(0,'g');right();
   }
   keyin('}');
   if ( command_state() || cfg==CFG_STRING || p_window_state:=='I' ||
      p_SyntaxIndent<0 || p_indent_style!=INDENT_SMART ||
      _in_comment() ) {
   } else if (_argument=='') {
      _str line;
      get_line(line);
      if (line=='}') {
         int col=c_endbrace_col(be_style);
         if (col) {
            replace_line(indent_string(col-1):+'}');
            p_col=col+1;
         }
      }
      _undo('S');
   }
}
/**
 * look for beginning of statement by searching for the following
 * <PRE>
 *      '{', '}', ';', ':', 'if', 'while','switch','for', 'with' (perl)
 * </PRE>
 * <P>
 * If a non-alpha symbol is found, we look ahead for the first a non-blank
 * character that is not in a comment.
 * <P>
 * NOTE:  Calling this function for code like the following will
 *        find the beginning of the code block and not the statement:
 * <PRE>
 *    <Find Here>for (...) ++i<Cursor Here>
 *    <Find Here>if/while (...) ++i<Cursor Here>
 * </PRE>
 *
 * @param RestorePos
 * @param SkipFirstHit
 * @param ReturnFirstNonBlank
 * @param FailIfNoPrecedingText
 * @param AlreadyRecursed
 * @param FailWithMinus1_IfNoTextAfterCursor
 *
 * @return int
*/
int c_begin_stat_col(boolean RestorePos,boolean SkipFirstHit,boolean ReturnFirstNonBlank,
                     boolean FailIfNoPrecedingText=false,
                     boolean AlreadyRecursed=false,
                     boolean FailWithMinus1_IfNoTextAfterCursor=false)
{
   int orig_linenum=p_line;
   int orig_col=p_col;
   //ReturnCurColIfCursorBetweenOpenBraceAndEOF=1;
   typeless p;
   save_pos(p);
   int status=search('[{};:()]|with|elsif|elseif|if|while|lock|switch|for|foreach|using','-R@');
   int nesting=0;
   boolean hit_top=false;
   int MaxSkipPreprocessing=VSCODEHELP_MAXSKIPPREPROCESSING;
   for (;;) {
      if (status) {
         top();
         hit_top=true;
      } else {
         int cfg=_clex_find(0,'g');
         if (cfg==CFG_COMMENT || cfg==CFG_STRING) {
            SkipFirstHit=false;
            status=repeat_search();
            continue;
         }
         switch (get_text()) {
         case '(':
            FailIfNoPrecedingText=false;
            if (nesting>0) {
               --nesting;
            }
            SkipFirstHit=false;
            status=repeat_search();
            continue;
         case ')':
            FailIfNoPrecedingText=false;
            ++nesting;
            SkipFirstHit=false;
            status=repeat_search();
            continue;
         }
         if (SkipFirstHit || nesting) {
            FailIfNoPrecedingText=false;
            SkipFirstHit=false;
            status=repeat_search();
            continue;
         }
         if (_in_c_preprocessing()) {
            --MaxSkipPreprocessing;
            if (MaxSkipPreprocessing<=0) {
               status=STRING_NOT_FOUND_RC;
               continue;
            }
            SkipFirstHit=false;
            begin_line();
            status=repeat_search();
            continue;
         }

         _str ch=get_text();
         if (!AlreadyRecursed && ch:==':') {
            typeless p2;
            save_pos(p2);
            if (p_col!=1) {
               left();
               // IF we are seeing  classname::name
               if (get_text()==':') {
                  status=repeat_search();
                  continue;
               }
               right();
            }
            int col=c_begin_stat_col(false,true,false,false,1);
            _str junk,word=cur_word(junk);
            if (word=='case' || word=='default') {
               restore_pos(p2);
               right();
            } else if (!ext_inherits_from("cs") && (word=='public' || word=='private' || word=='protected') || 
                       (ext_inherits_from("c") && word=='signals')
                       ) {
               restore_pos(p2);
               right();
            }
         } else {
            /*
                Handle where constraint case for csharp.  Need to go back to beginning of class definition.
                The only down side to doing this is that if the constraints are on multiple lines we will
                indent back to the "where" column.  This is not a likely case so we can forget about it.

                class myclass<a>
                    where a: constraint1,constraint2,constraint3
                    where b: constraint1,constraint2,constraint3
            */
            if (AlreadyRecursed && ch:==':' && ext_inherits_from('cs')) {
               _str line, word, more;
               get_line(line);
               parse line with word more':';
               if (word=='where') {
                  status=repeat_search();
                  continue;
               }
            }
            if (isalpha(ch)) {
               if(cfg!=CFG_KEYWORD) {
                  if (cfg!=CFG_STRING && cfg!=CFG_COMMENT) {
                     FailIfNoPrecedingText=false;
                  }
                  status=repeat_search();
                  continue;
               }
            } else {
               right();
            }
         }
      }
      status=_clex_skip_blanksNpp();
      if (status) {
         restore_pos(p);
         /*
             Would could have an open brace followed by blanks and eof.
         */
         if (!hit_top) {
            if (!FailWithMinus1_IfNoTextAfterCursor) {
               return(p_col);
            }
            return(-1);
         }
         return(0);
      }
      /*
          We could have the following:

            class name:public name2 {

          recurse to look for "case" keyword

      */
      if (ReturnFirstNonBlank) {
         first_non_blank();
      }
      int col=p_col;
      if (hit_top && FailIfNoPrecedingText && (p_line>orig_linenum || (p_line==orig_linenum)&& p_col>orig_col)) {
         return(0);
      }
      if (RestorePos) {
         restore_pos(p);
      }
      return(col);
   }

}

/**
 * On entry, the cursor is sitting on a } (close brace)
 * <PRE>
 * static void
 *    main () /* this is a test */ {
 * }
 * static void main /* this is a test */
 *   ()
 * {
 * }
 * </PRE>
 *
 * @param be_style  begin-end brace style
 * <PRE>
 * for (;;) {     for (;;)        for (;;)
 *                {                  {
 *                }                  }
 * }
 * style 0        style 1         style 2
 * </PRE>
 *
 * @return
 * Returns column where end brace should go.
 * Returns 0 if this function does not know the column where the
 * end brace should go.
*/
int c_endbrace_col(int be_style)
{
   boolean style3_MustBackIndent=false;
   return(c_endbrace_col2(be_style,style3_MustBackIndent));
}
int c_endbrace_col2(int be_style,boolean &style3_MustBackIndent)
{
   style3_MustBackIndent=false;
   if (p_lexer_name=='') {
      return(0);
   }
   typeless p;
   save_pos(p);
   --p_col;
   // Find matching begin brace
   int status=_find_matching_paren(def_pmatch_max_diff);
   if (status) {
      restore_pos(p);
      return(0);
   }
   // Assume end brace is at level 0
   if (p_col==1) {
      restore_pos(p);
      return(1);
   }
   typeless p2;
   save_pos(p2);
   int begin_brace_col=p_col;
   // Check if the first char before open brace is close paren
   int col= find_block_col();
   if (!col) {
      restore_pos(p2);
      if (_isVarInitList(true)) {
         restore_pos(p2);
         first_non_blank();
         col=p_col;
         restore_pos(p);
         return(col);
      }
      restore_pos(p2);
#if 0
      if ((be_style & VS_C_OPTIONS_STYLE2_FLAG) && def_style3_indent_all_braces) {
         // check if this parenthesis is on a line by itself;
         get_line(line);
         if (line=="{") {
            style3_MustBackIndent=true;
            first_non_blank();
            col=p_col;
            restore_pos(p);
            return(col);
         }
      }
#endif
      col=c_begin_stat_col(true,true,true);
      restore_pos(p);
      if ((be_style & VS_C_OPTIONS_STYLE2_FLAG) && def_style3_indent_all_braces) {
         style3_MustBackIndent=true;
         parse name_info(_edit_window().p_index) with . .;
         col+=p_SyntaxIndent;
      }
      return(col);
   }
   style3_MustBackIndent=true;
   int style=(be_style & (VS_C_OPTIONS_STYLE1_FLAG|VS_C_OPTIONS_STYLE2_FLAG));
   if ((be_style & (VS_C_OPTIONS_STYLE2_FLAG))) {
      restore_pos(p);
      //return(begin_brace_col);
      parse name_info(_edit_window().p_index) with . . ;
      return(col+p_SyntaxIndent);
   }
   restore_pos(p);
   return(col);
}
/**
 * Starts from cursor position and skips over spaces, tabs, and comments.
 *
 * @param options may contain one or more of the following option letters:
 * <pre>
 *    -	Search backwards
 *    m	Search within selection
 *    h	Search through hidden lines.
 * </pre>
 * @return	Returns 0 if non-blank character is found, nonzero otherwise.
 * If this functions fails the cursor is moved but its final location may
 * not be the top or bottom of the buffer (we need to change this should
 * be more concrete).
 * 	
 * @see _clex_find
 * @see p_color_flags
 * @see _clex_load
 *
 * @categories Edit_Window_Methods, Editor_Control_Methods, Search_Functions
 */
int _clex_skip_blanks(_str options='')
{
   _str search_options='@r':+options;
   _str clex_options='n'stranslate(lowcase(options),'','m');
   clex_options=stranslate(clex_options,'','h');
   for (;;) {
      // search for non-blank character
      int status=search('[~ \t\r\n]',search_options);
      if (status) return(1);
      typeless orig_values;
      int embedded_status=_EmbeddedStart(orig_values,'');
      if (_in_comment()) {
         status=_clex_find(COMMENT_CLEXFLAG,clex_options);
         if (status) {
            /* This changes was made for select_proc() */
            if (pos('-',options)) {
               top();
            } else {
               bottom();
            }
            if( embedded_status==1 ) _EmbeddedEnd(orig_values);
            return(1);
         }
         if( embedded_status==1 ) _EmbeddedEnd(orig_values);
         continue;
      }
      if( embedded_status==1 ) _EmbeddedEnd(orig_values);
      return(0);
   }

}

/**
 * Translates the reults of a _clex_find(0,'g') into one of the
 * CFG_<> constants that can be used in a subsequent _clex_find() call.
 *
 * @param clexflags
 *
 * @return
 */
int _clex_translate(int clexflags)
{
   switch (clexflags) {
   case CFG_KEYWORD:
      return(KEYWORD_CLEXFLAG);
   case CFG_LINENUM:
      return(LINENUM_CLEXFLAG);
   case CFG_NUMBER:
      return(NUMBER_CLEXFLAG);
   case CFG_STRING:
      return(STRING_CLEXFLAG);
   case CFG_COMMENT:
      return(COMMENT_CLEXFLAG);
   case CFG_PPKEYWORD:
      return(PPKEYWORD_CLEXFLAG);
   case CFG_PUNCTUATION:
      return(SYMBOL1_CLEXFLAG);
   case CFG_LIBRARY_SYMBOL:
      return(SYMBOL2_CLEXFLAG);
   case CFG_OPERATOR:
      return(SYMBOL3_CLEXFLAG);
   case CFG_USER_DEFINED:
      return(SYMBOL4_CLEXFLAG);
   case CFG_FUNCTION:
      return(FUNCTION_CLEXFLAG);
   default :
      return(OTHER_CLEXFLAG);
   }
}

/**
 * Given the current color coding element, this function places
 * the cursor on the first character of that element.
 *
 * @return Returns 0 if successful.
 */
int _clex_find_start()
{
   typeless p; save_pos(p);
   int clexflags=_clex_find(0,'g');
   if (!clexflags) {
      return(1);
   }
   clexflags = _clex_translate(clexflags);
   int status=_clex_find(clexflags,'n-');
   if (status) {
      top();
   }
   status=_clex_find(clexflags);
   if (status) {
      // this should not happen, but just in case
      restore_pos(p);
      return(1);
   }
   return(0);
}

/**
 * Given the current color coding element, this function places the cursor on the
 * last character of that element.
 *
 * @return Returns 0 if successful.
 */
int _clex_find_end()
{
   typeless p;
   save_pos(p);
   int clexflags=_clex_find(0,'g');
   if (!clexflags) {
      return(1);
   }
   clexflags = _clex_translate(clexflags);
   int status=_clex_find(clexflags,'n');
   if (status) {
      bottom();
   }
   status=_clex_find(clexflags,'o-');
   if (status) {
      // this should not happen, but just in case
      restore_pos(p);
      return(1);
   }
   return(0);
}

static int find_block_col()
{
   _str word;
   int col=0;
   --p_col;
   if (_clex_skip_blanks('-')) return(0);
   if (get_text()!=')') {
      if (_clex_find(0,'g')!=CFG_KEYWORD) {
         return(0);
      }
      word=cur_word(col);
      if (word=='do' || word=='else') {
         first_non_blank();
         return(p_col);
         //return(p_col-length(word)+1);
      }
      return(0);
   }
   // Here we match round parens. ()
   int status=_find_matching_paren(def_pmatch_max_diff);
   if (status) return(0);
   if (p_col==1) return(1);
   --p_col;

   if (_clex_skip_blanks('-')) return(0);
   /*if (_clex_find(0,'g')!=CFG_KEYWORD) {
      return(0);
   }
   */
   word=cur_word(col);
   if (pos(' 'word' ',' with for foreach if elsif elseif switch while lock using ')) {
      first_non_blank();
      return(p_col);
      //return(p_col-length(word)+1);
   } else if (ext_inherits_from('java')) {
      // Check if we have a new construct
      p_col=_text_colc(col,'I');
      if (p_col>1) {
         left();
         if (_clex_skip_blanks('-')) return(0);
         word=cur_word(col);
         if (word=='new') {
            p_col=_text_colc(col,'I');
            col=p_col;
            first_non_blank();
            if (col!=p_col) {
               p_col+=p_SyntaxIndent;
            }
            return(p_col);
         }
      }
   }
   return(0);
}
#define EXPAND_WORDS (' #define #elif #else #endif #error #if #ifdef #ifndef':+\
                ' #include #pragma #undef else enum typedef static struct class' :+\
                ' union public private protected ')

#define JAVA_ONLY_EXPAND_WORDS ' final import int interface package '
#define CS_ONLY_EXPAND_WORDS ' #region #endregion '
#define IDL_ONLY_EXPAND_WORDS ' attribute in inout '
#define PHP_ONLY_EXPAND_WORDS ' import '

static SYNTAX_EXPANSION_INFO cpp_space_words:[] = {
   '#define'      => { "#define" },
   '#elif'        => { "#elif" },
   '#else'        => { "#else" },
   '#endif'       => { "#endif" },
   '#error'       => { "#error" },
   '#if'          => { "#if" },
   '#ifdef'       => { "#ifdef" },
   '#ifndef'      => { "#ifndef" },
   '#include'     => { "#include" },
   '#pragma'      => { "#pragma" },
   '#undef'       => { "#undef" },
   'break'        => { "break;" },
   'case'         => { "case" },
   'class'        => { "class" },
   'continue'     => { "continue;" },
   'default'      => { "default:" },
   'do'           => { "do { ... } while ( ... );" },
   'else'         => { "else { ... }" },
   'enum'         => { "enum" },
   'else if'      => { "else if ( ... ) { ... }" },
   'for'          => { "for ( ... ) { ... }" },
   'if'           => { "if ( ... ) { ... }" },
   'main'         => { "main(int argc, char *argv[]) { ... }" },
   'printf'       => { "printf(\"" },
   'private'      => { "private:" },
   'protected'    => { "protected:" },
   'public'       => { "public:" },
   'return'       => { "return" },
   'struct'       => { "struct" },
   'switch'       => { "switch ( ... ) { ... }" },
   'try'          => { "try { ... } catch ( ... ) { ... }" },
   'typedef'      => { "typedef" },
   'union'        => { "union" },
   'while'        => { "while ( ... ) { ... }" },
};

static _str else_space_words[] = { 'else', 'else if' };

static SYNTAX_EXPANSION_INFO cs_space_words:[] = {
   '#define'      => { "#define" },
   '#elif'        => { "#elif" },
   '#else'        => { "#else" },
   '#endif'       => { "#endif" },
   '#endregion'   => { "#endregion" },
   '#error'       => { "#error" },
   '#if'          => { "#if" },
   '#ifdef'       => { "#ifdef" },
   '#ifndef'      => { "#ifndef" },
   '#include'     => { "#include" },
   '#pragma'      => { "#pragma" },
   '#region'      => { "#region" },
   '#undef'       => { "#undef" },
   'break'        => { "break;" },
   'case'         => { "case" },
   'class'        => { "class" },
   'continue'     => { "continue;" },
   'default'      => { "default:" },
   'do'           => { "do { ... } while ( ... );" },
   'else'         => { "else { ... }" },
   'enum'         => { "enum" },
   'else if'      => { "else if ( ... ) { ... }" },
   'finally'      => { "finally { ... }" },
   'fixed'        => { "fixed" },
   'for'          => { "for ( ... ) { ... }" },
   'foreach'      => { "foreach ( ... ) { ... }" },
   'if'           => { "if ( ... ) { ... }" },
   'int'          => { "int" },
   'interface'    => { "interface" },
   'lock'         => { "lock ( ... ) { ... }" },
   'main'         => { "public static void Main (string []args) { ... }" },
   'private'      => { "private" },
   'protected'    => { "protected" },
   'public'       => { "public" },
   'return'       => { "return" },
   'struct'       => { "struct" },
   'switch'       => { "switch ( ... ) { ... }" },
   'try'          => { "try { ... } catch ( ... ) { ... }" },
   'union'        => { "union" },
   'using'        => { "using" },
   'while'        => { "while ( ... ) { ... }" },
};

static SYNTAX_EXPANSION_INFO java_space_words:[] = {
   '#define'      => { "#define" },
   '#elif'        => { "#elif" },
   '#else'        => { "#else" },
   '#endif'       => { "#endif" },
   '#error'       => { "#error" },
   '#if'          => { "#if" },
   '#ifdef'       => { "#ifdef" },
   '#ifndef'      => { "#ifndef" },
   '#include'     => { "#include" },
   '#pragma'      => { "#pragma" },
   '#undef'       => { "#undef" },
   'break'        => { "break;" },
   'case'         => { "case" },
   'class'        => { "class" },
   'continue'     => { "continue;" },
   'default'      => { "default:" },
   'do'           => { "do { ... } while ( ... );" },
   'else'         => { "else { ... }" },
   'enum'         => { "enum" },
   'for'          => { "for ( ... ) { ... }" },
   'if'           => { "if ( ... ) { ... }" },
   'int'          => { "int" },
   'interface'    => { "interface" },
   'main'         => { "public static void main (String args[]) { ... }" },
   'package'      => { "package" },
   'private'      => { "private" },
   'protected'    => { "protected" },
   'public'       => { "public" },
   'return'       => { "return" },
   'switch'       => { "switch ( ... ) { ... )" },
   'try'          => { "try { ... } catch ( ... ) { ... }" },
   'while'        => { "while ( ... ) { ... }" },
   'else if'      => { "else if ( ... ) { ... }" },
   'static'       => { "static" },
   '@interface'   => { "@interface" },
   'final'        => { "final" },
   'finally'      => { "finally" },
   'import'       => { "import" },
};

static SYNTAX_EXPANSION_INFO javascript_space_words:[] = {
   'break'        => { "break;" },
   'case'         => { "case" },
   'continue'     => { "continue;" },
   'default'      => { "default:" },
   'do'           => { "do { ... } while ( ... );" },
   'else'         => { "else { ... }" },
   'else if'      => { "else if ( ... ) { ... }" },
   'finally'      => { "finally { ... }" },
   'for'          => { "for ( ... ) { ... }" },
   'if'           => { "if ( ... ) { ... }" },
   'return'       => { "return" },
   'switch'       => { "switch ( ... ) { ... }" },
   'while'        => { "while ( ... ) { ... }" },
   'export'       => { "export" },
   'function'     => { "function" },
   'import'       => { "import" },
   'try'          => { "try { ... } catch ( ... ) { ... }" },
   'var'          => { "var" },
   'with'         => { "with ( ... ) { ... }" },
};

static SYNTAX_EXPANSION_INFO php_space_words:[] = {
   'break'        => { "break;" },
   'case'         => { "case" },
   'class'        => { "class" },
   'continue'     => { "continue;" },
   'default'      => { "default:" },
   'do'           => { "do { ... } while ( ... );" },
   'else'         => { "else { ... }" },
   'for'          => { "for ( ... ) { ... )" },
   'if'           => { "if ( ... ) { ... }" },
   'return'       => { "return" },
   'switch'       => { "switch ( ... ) { ... }" },
   'while'        => { "while ( ... ) { ... }" },
   'function'     => { "function" },
   'import'       => { "import" },
   'var'          => { "var" },
   'elseif'       => { "elseif ( ... ) { ... }" },
};

static SYNTAX_EXPANSION_INFO idl_space_words:[] = {
   'attribute'    => { "attribute" },
   'case'         => { "case" },
   'default'      => { "default:" },
   'exception'    => { "exception ... { ... }" },
   'in'           => { "in" },
   'inout'        => { "inout" },
   'interface'    => { "interface ... { ... }" },
   'module'       => { "module ... { ... }" },
   'sequence'     => { "sequence < ... >" },
   'struct'       => { "struct ... { ... }" },
   'switch'       => { "switch ( ... ) { ... }" },
   'typedef'      => { "typedef" },
   'union'        => { "union ... { ... }" },
};

_str _skip_pp;

int c_get_info(var Noflines,var cur_line,var first_word,var last_word,
               var rest,var non_blank_col,var semi,var prev_semi,
               boolean in_smart_paste=false)
{
   typeless old_pos;
   save_pos(old_pos);
   first_word='';last_word='';non_blank_col=p_col;
   int orig_col=p_col;
   int i,j;
   for (j=0;  ; ++j) {
      get_line_raw(cur_line);
      boolean bool = false;
      if (in_smart_paste) {
         _begin_line();
         i=verify(cur_line,' '\t);
         if ( i ) p_col=text_col(cur_line,i,'I');
         bool=cur_line!='' && (substr(strip(cur_line),1,1)!='#' || _skip_pp=='') && _clex_find(0,'g')!=CFG_COMMENT;
      } else {
         bool=cur_line!='' && (substr(strip(cur_line),1,1)!='#' || _skip_pp=='');
      }
      if ( bool){
         _str line;
         _str before_brace;
         parse cur_line with line '/*',p_rawpos; /* Strip comment on current line. */
         parse line with line '//',p_rawpos; /* Strip comment on current line. */
         parse line with before_brace '{',p_rawpos +0 last_word;
         parse strip(line,'L') with first_word '[({:; \t]',(p_rawpos'r') +0 rest;
         last_word=strip(last_word);
         typeless expand, be_style, indent_fl,indent_case;
         parse name_info(_edit_window().p_index) with . expand . . be_style indent_fl . indent_case .;
         int syntax_indent=p_SyntaxIndent;
         if (last_word=='{' && !(be_style & VS_C_OPTIONS_STYLE2_FLAG)) {
            typeless p2;
            save_pos(p2);
            p_col=text_col(before_brace);
            _clex_skip_blanks('-');
            int status=1;
            if (get_text()==')') {
               status=_find_matching_paren(def_pmatch_max_diff);
            }
            if (!status) {
               status=1;
               if (p_col==1) {
                  up();_end_line();
               } else {
                  left();
               }
               _clex_skip_blanks('-');
               if (_clex_find(0,'g')==CFG_KEYWORD) {
                  _str junk;
                  _str kwd=cur_word(junk);
                  status=(int) !pos(' 'kwd' ',' with if elsif elseif while switch for foreach lock using ');
               }
            }
            if (status) {
               non_blank_col=text_col(line,pos('[~ \t]|$',line,1,p_rawpos'r'),'I');
               restore_pos(p2);
            } else {
               get_line_raw(line);
               non_blank_col=text_col(line,pos('[~ \t]|$',line,1,p_rawpos'r'),'I');
               /* Use non blank of start of if, do, while, which, or for. */
            }
         } else {
            non_blank_col=text_col(line,pos('[~ \t]|$',line,1,p_rawpos'r'),'I');
         }
         Noflines=j;
         break;
      }
      if ( up() ) {
         restore_pos(old_pos);
         return(1);
      }
      if (j>=100) {
         restore_pos(old_pos);
         return(1);
      }
   }
   if (in_smart_paste) {
      if (!j) p_col=orig_col;
   }
   typeless p='';
   if ( j ) {
      p=1;
   }
   semi=stat_has_semi(p);
   prev_semi=prev_stat_has_semi();
   restore_pos(old_pos);
   return(0);
}
boolean _in_c_preprocessing()
{
   typeless p;
   save_pos(p);
   //get_line(line);line=strip(line,'L');
   for (;;) {
      _str line;
      get_line(line);line=strip(line,'L');
      if (substr(line,1,1)=="#") {
         restore_pos(p);
         return(true);
      }
      up();
      if (_on_line0()) {
         restore_pos(p);
         return(false);
      }
      //get_line(line);line=strip(line,'L');
      _end_line();left();
      if (get_text()=='\') {
      //if (last_char(line)=='\') {
         _end_line();left();
         int cfg=_clex_find(0,'g');
         if (cfg==CFG_COMMENT && cfg==CFG_STRING) {
            restore_pos(p);
            return(false);
         }
      } else {
         restore_pos(p);
         return(false);
      }
   }

}
/**
 * @return non-zero number if pass through to enter key required
 */
_str _c_expand_enter(int syntax_indent,int expand,int be_style,
                     int indent_fl,int indent_case)
{
   typeless p;
   save_pos(p);
   int orig_linenum=p_line;
   int orig_col=p_col;
   _str enter_cmd=name_on_key(ENTER);
   if (enter_cmd=='nosplit-insert-line') {
      _end_line();
   }
   if (_in_c_preprocessing()) {
      restore_pos(p);
      return(1);
   }
   int begin_col=c_begin_stat_col(false /* No RestorePos */,
                                  false /* Don't skip first begin statement marker */,
                                  false /* Don't return first non-blank */,
                                  1  /* Return 0 if no code before cursor. */,
                                  false,
                                  1
                                  );
   if (!begin_col /*|| (p_line>orig_linenum)*/) {
      restore_pos(p);
      return(1);
   }
   int status=0;
   int style1=be_style & VS_C_OPTIONS_STYLE1_FLAG;
   int style2=be_style & VS_C_OPTIONS_STYLE2_FLAG;
   boolean LineEndsWithBrace=false;
   int java=0;
   if (ext_inherits_from('java') || ext_inherits_from('cs')) {
      java=1;
   } else if (ext_inherits_from('js') || ext_inherits_from('cfscript')) {
      java=2;
   }
   if (p_line>orig_linenum) {
      if (p_col==1) {
         up();_end_line();
      } else {
         left();
      }
      _clex_skip_blanksNpp("-");
      LineEndsWithBrace= (orig_linenum==p_line && get_text()=='{');
      first_non_blank();
   } else if (p_line==orig_linenum && begin_col<0) {
      LineEndsWithBrace= (orig_linenum==p_line && get_text()=='{');
      first_non_blank();
   }
   int col=0,first_word_col=p_col;
   _str junk, first_word=cur_word(junk);
   int first_word_color=_clex_find(0,'g');
   _str cur_line, line, orig_line;
   get_line_raw(cur_line);
   boolean BeginningOfStatementOnSameLine=(orig_linenum==p_line);
   restore_pos(p);
   enter_cmd=name_on_key(ENTER);
   if ( expand && BeginningOfStatementOnSameLine &&
        !(_expand_tabsc(orig_col)!="" &&
          (_will_split_insert_line()
          )
        ) &&
        first_word_color==CFG_KEYWORD
        ) {
      if ( cur_line=='main' && !java) {
         status=c_insert_main();
      } else if ( first_word=='for' && name_on_key(ENTER):=='nosplit-insert-line' ) {
         if ( name_on_key(ENTER):=='nosplit-insert-line' ) {
            /* tab to fields of C for statement */
            p_col=orig_col;
            line=expand_tabs(cur_line);
            int semi1_col=pos(';',line,p_col,p_rawpos);
            if ( semi1_col>0 && semi1_col>=p_col ) {
               p_col=semi1_col+1;
            } else {
               int semi2_col=pos(';',line,semi1_col+1,p_rawpos);
               if ( (semi2_col>0) && (semi2_col>=p_col) ) {
                  p_col=semi2_col+1;
               } else {
                  status=1;
               }
            }
         } else {
            status=1;
         }
      } else if ( (first_word=='case' || first_word=='default') &&
                 (orig_col>first_word_col ||
                  enter_cmd=='nosplit-insert-line') ) {
         _str eol='';
         if( _will_split_insert_line() ){
            get_line_raw(orig_line);
            eol=expand_tabs(orig_line,p_col,-1,'s');
            replace_line_raw(expand_tabs(orig_line,1,p_col-1,'s'));
         }
         /* Indent case based on indent of switch. */
         col=_c_last_switch_col();
         if ( col && eol:=='') {
            if ((indent_case && indent_case!='') || (be_style&VS_C_OPTIONS_STYLE2_FLAG)) {
               col=col+syntax_indent;
            }
            replace_line_raw(indent_string(col-1):+""strip(cur_line,'L'));
            _end_line();
         }
         indent_on_enter(syntax_indent);
         if (eol:!='') {
            replace_line_raw(indent_string(p_col-1):+eol);
         }
      } else if ( first_word=='switch' && LineEndsWithBrace) {
         down();
         get_line_raw(line);
         up();
         int extra_case_indent=0;
         if ((indent_case && indent_case!='') || (be_style&VS_C_OPTIONS_STYLE2_FLAG)) {
            extra_case_indent=syntax_indent;
         }
         if ( pos('}',line,1,p_rawpos) > 0 ) {
            indent_on_enter(syntax_indent);
            get_line_raw(line);
            if ( line=='' ) {
               col=p_col-syntax_indent;
               replace_line_raw(indent_string(col-1+extra_case_indent)'case ');
               _end_line();
               c_maybe_list_args(true);
            }
         } else {
            indent_on_enter(syntax_indent);
            get_line_raw(line);
            if ( line=='' ) {
               col=p_col-syntax_indent;
               replace_line_raw(indent_string(col-1+extra_case_indent)'case ');
               _end_line();
               c_maybe_list_args(true);
            }
         }
     } else {
       status=1;
     }
   } else {
     status=1;
   }
   if ( status ) {  /* try some more? Indenting only. */
      status=0;
      col=c_indent_col2(0,0);
      indent_on_enter('',col);
   }
   return(status);
}
/**
 * @return Return column position on switch or 0 if not found
 */
int _c_last_switch_col()
{
   if (p_lexer_name=='') {
      return(0);
   }
   typeless p;
   save_pos(p);
   // Find switch at same brace level
   // search for begin brace,end brace, and switch not in comment or string
   int status=search('\{|\}|switch','@r-');
   int level=0;
   for (;;) {
      if (status) {
         restore_pos(p);
         return(0);
      }
      _str word=get_match_text();
      int color=_clex_find(0,'g');
      //messageNwait('word='word);
      if (color!=CFG_STRING && color!=CFG_COMMENT) {
         switch (word) {
         case '}':
            --level;
            break;
         case '{':
            ++level;
            break;
         default:
            if (color==CFG_KEYWORD && level== 1) {
               int result=p_col;
               restore_pos(p);
               return(result);
            }
         }
      }
      status=repeat_search();
   }
}

static int NoSyntaxIndentCase(int non_blank_col,int orig_linenum,int orig_col,typeless p,int syntax_indent)
{
   //_message_box("This case not handled yet");
   // Smart paste should set the non_blank_col
   if (non_blank_col) {
      //messageNwait("fall through case 1");
      restore_pos(p);
      return(non_blank_col);
   }
   restore_pos(p);
   int begin_stat_col=c_begin_stat_col(false /* No RestorePos */,
                                       false /* Don't skip first begin statement marker */,
                                       true  /* Don't return first non-blank */
                                       );

   if (begin_stat_col && (p_line<orig_linenum ||
                          (p_line==orig_linenum && p_col<=orig_col)
                         )
      ) {
#if 0
      /*
          We could have code at the top of a file like the following:

             int myproc(int i)<ENTER>

             int myvar=<ENTER>
             class foo :<ENTER>
                public name2

      */
      //messageNwait("fall through case 2");
      restore_pos(p);
      return(begin_stat_col);
#endif
      /*
         Check if partial statement ends with close paren.  This
         could be a function declaration.

         Another to handle this is to to indent any way and then
         move the open brace to the correct colmun position when
         the users types it.
      */
      typeless p2;
      save_pos(p2);
      p_line=orig_linenum;p_col=orig_col;
      if (p_col==1) {
         up();_end_line();
      } else {
         left();
      }
      int status=_clex_skip_blanksNpp("-");
      if (status) {
         restore_pos(p);
         return(orig_col);
      }
      _str ch=get_text();
      if (ch:==")") {
         restore_pos(p);
         return(begin_stat_col);
      }
      restore_pos(p2);
      /*
         Here we have something like
         int i;
            int k,<ENTER>
               <Cursor goes here>
               OR
         VOID<ENTER>
         <Cursor goes here>myproc()
      */
      int col=p_col;
      // Here we assume that functions start in column 1 and
      // variable declarations or statement continuations do not.
      // This seems to be a common solution.
      if (p_col==1 && ch!=',') {
         restore_pos(p);
         return(col);
      }
      int nextline_indent=syntax_indent;
      restore_pos(p);
      return(col+nextline_indent);
   }
   restore_pos(p);
   _str line;
   get_line(line);
   line=expand_tabs(line);
   if (line=="") {
      restore_pos(p);
      return(p_col);
   }
   //messageNwait("fall through case 3");
   first_non_blank();
   int col=p_col;
   restore_pos(p);
   return(col);
}
/**
 * Skip blanks and preprocessing
 *
 * @param options    search options for {@link _clex_find} and {@link search()}
 *
 * @return 0 on success, nonzero otherwise
 */
int _clex_skip_blanksNpp(_str options='')
{
   int MaxSkipPreprocessing=VSCODEHELP_MAXSKIPPREPROCESSING;
   int backwards=pos('-',options);
   for (;;) {
      int status=_clex_skip_blanks(options);
      if (status) {
         return(status);
      }
      /*if (p_line>FailIfPastLinenum) {
         messageNwait("p_line="p_line" FailIfPastLinenum="FailIfPastLinenum);
         return(STRING_NOT_FOUND_RC);
      }*/
      if (!_in_c_preprocessing()) {
         return(status);
      }
      --MaxSkipPreprocessing;
      if (MaxSkipPreprocessing<=0) {
         return(STRING_NOT_FOUND_RC);
      }
      if (backwards) {
         up();_end_line();
      } else {
         _end_line();
      }
   }
}
static int HandlePartialStatement(int statdelim_linenum,
                                  int sameline_indent,
                                  int nextline_indent,
                                  int orig_linenum,int orig_col)
{
   _str orig_ch=get_text();
   typeless orig_pos;
   save_pos(orig_pos);
   //linenum=p_line;col=p_col;

   /*
       Note that here we don't return first non-blank to handle the
       following case:

       for (;
            ;<ENTER>) {

       However, this does effect the following unusual case
           if (i<j) {abc;<ENTER>def;
           <end up here which is not correct>

       We won't worry about this case because it is unusual.
   */
   int begin_stat_col=c_begin_stat_col(false /* No RestorePos */,
                                       false /* Don't skip first begin statement marker. */,
                                       false /* Don't return first non-blank */,
                                       false,
                                       false,
                                       1   // Fail if no text after cursor
                                       );
   if (begin_stat_col>0 && (p_line<orig_linenum || (p_line==orig_linenum && p_col<orig_col))
        /* && (linenum!=p_line || col!=p_col) */
      ) {
      // Now get the first non-blank column.
      begin_stat_col=c_begin_stat_col(false /* No RestorePos */,
                                      false /* Don't skip first begin statement marker. */,
                                      true /* Return first non-blank */
                                      );
      /*
         Check if partial statement ends with close paren.  This
         could be a function declaration.

         Another to handle this is to to indent any way and then
         move the open brace to the correct colmun position when
         the users types it.
      */
      typeless p;
      save_pos(p);
      p_line=orig_linenum;p_col=orig_col;
      if (p_col==1) {
         up();_end_line();
      } else {
         left();
      }
      _clex_skip_blanksNpp("-");
      _str ch=get_text();
      if (ch:==")") {
         return(begin_stat_col);
      }
      if (orig_ch:=='}' && ch:==',' && statdelim_linenum==p_line) {
         /*
             Also check if this line ends with a comma and handle the
             case where the user is in a declaration list like
             the following.

             MYSTRUCT array[]={
                {a,b,c},<ENTER>
                a,
                {a,b,c},<ENTER>
                {a,b,c},a,b,<ENTER>
                {a,{a,b,c}},<ENTER>
                d,
                {a,
                 {a,b,c}},
                 x,<ENTER>
                },
                b,
         */
         restore_pos(orig_pos);
         int status=_find_matching_paren(def_pmatch_max_diff);
         if (!status) {
            first_non_blank();
            return(p_col);
         }
      }
      restore_pos(p);
      /*
         IF semicolon is on same line as extra characters

         Example
            {b=<ENTER>
      */
      if (p_line==statdelim_linenum) {
         return(begin_stat_col+sameline_indent);
      }
      /*
         Here we have something like
         int i;
            int k,<ENTER>
               <Cursor goes here>
               OR
         VOID<ENTER>
         <Cursor goes here>myproc()
      */
      int col=p_col;
      // Here we assume that functions start in column 1 and
      // variable declarations or statement continuations do not.
      // This seems to be a common solution.
      if (p_col==1 && ch!=',') {
         return(col);
      }
      return(col+nextline_indent);
   }
   return(0);
}
/**
 * NOTE:  The caller should check if the user is calling this
 * function when inside a comment (use _in_comment(1) function).
 *
 * @param non_blank_col        All parameters are ignored except
 *                             non_blank_col.  Specify non_blank_col==0
 *                             if you want it ignored too.
 * @param pasting_open_brace2  ignored
 *
 * @return indent column
*/
int c_indent_col(int non_blank_col,boolean pasting_open_brace2)
{
   return(c_indent_col2(non_blank_col,pasting_open_brace2));
}
static boolean _isQmarkExpression()
{
   // cursor is sitting colon
   /*
      could have
                (c)?s:t,
                MYCLASS():a(1),b(2) {
             class name1:public<ENTER>

            default :
            case 'a':
            case ('a'+'b')-1:
            public:
            private:
            protected:
      Give up on } for now.
   */
   int status=search('[?;})]|class|default|case|public|private|protected','-@rxcs');
   for (;;) {
      if (status) {
         return(false);
      }
      switch(get_match_text()) {
      case '?':
         return(true);
      case ';':
      case '}':
         return(false);
      case ')':
         status=find_matching_paren();
         if (status) {
            return(false);
         }
         status=repeat_search();
         continue;
      default:
         if (_clex_find(0,'g')==CFG_KEYWORD) {
            status=repeat_search();
            continue;
         }
         return(false);
      }
   }
}
static boolean _isVarInitList(boolean checkEnum=false)
{
   /*
      Check for the array/struct initialization case by
      check for equal sign before open brace.  This won't
      work if preprocessing is present.

        int array[]={
           a,
           b,<ENTER>
           int a,
           b,
           c,


      object array[]={
         "a","b",
         "c",{"a",
            "b","c"
         }
      };

      also check for enum declaration like
         enum {
           a,
           b,
      or
         enum XXX {
            a,
            b,


   */
   if (p_col==1) {
      up();_end_line();
   } else {
      left();
   }
   _clex_skip_blanksNpp('-');
   int cfg=_clex_find(0,'g');
   _str junk;
   boolean in_init=(get_text()=='=' || get_text()==']' ||
                    get_text()==',' || get_text()=='{' ||
                    (checkEnum && cfg==CFG_KEYWORD && cur_word(junk)=='enum'));
   //messageNwait('ch='get_text()' in_init='in_init);

   if (!in_init && checkEnum && cfg==CFG_WINDOW_TEXT &&
       pos('[a-zA-Z0-9_$]',get_text(),1,'r')) {

      int status=search("[^a-zA-Z0-9_$]","@-R");
      if (!status) {
         _clex_skip_blanksNpp('-');
         if (_clex_find(0,'g')==CFG_KEYWORD && cur_word(junk)=='enum') {
            in_init=1;
         }
      }
   }
   return(in_init);
}
static void parse_template_args()
{
   int nesting=0;
   for (;;) {
      if (c_sym_gtk()=='>') {
         ++nesting;
      } else if (c_sym_gtk()=='<') {
         --nesting;
         if (nesting<=0) {
            c_prev_sym();
            return;
         }
      } else if (c_sym_gtk()=='') {
         return;
      }
      c_prev_sym();
   }
}
static int c_indent_col2(int non_blank_col,boolean pasting_open_brace2)
{
   int orig_col=p_col;
   int orig_linenum=p_line;
   int orig_embedded=p_embedded;
   int col=orig_col;
   typeless p;
   save_pos(p);
   typeless expand, be_style, indent_fl, indent_case, UseContOnParameters;
   parse name_info(p_index) with . expand . . be_style indent_fl . indent_case UseContOnParameters .;
   int syntax_indent=p_SyntaxIndent;
   if ( syntax_indent<=0) {
      // Find non-blank-col
      return(NoSyntaxIndentCase(non_blank_col,orig_linenum,orig_col,p,0));
   }
   int style1=be_style & VS_C_OPTIONS_STYLE1_FLAG;
   int style2=be_style & VS_C_OPTIONS_STYLE2_FLAG;

   if (pasting_open_brace2) {
      // Look for for,while,switch
      typeless p2;
      save_pos(p2);
      col= find_block_col();
      restore_pos(p2);
      if (col) {
         restore_pos(p);
         if (style2) {
            return(col+syntax_indent);
         }
         return(col);
      }
      /*
          Note:
             pasting open brace does not yet work well
             for style2 when pasting brace out side class/while/for/switch blocks.
             Braces are not indented.

             pasting open brace does not yet work well
             for style2!=2 when pasting braces for a class.  Braces
             end up indented when they are not supposed to be.
      */
   }

   // locals
   int cfg=0;
   int begin_stat_col=0;
   _str ch='';
   _str word='';
   _str junk='';
   _str line='';
   _str kwd='';
   typeless p2,p3;

/*
   beginning of statement
     {,},;,:

   cases
     -  in-comment or in-string
     - for (;;) <ENTER>


     - myproc(myproc() <ENTER>
     - myproc(a,<ENTER>
     - myproc(a);
     - if/while/for/switch (...) <ENTER>
     - (col1)myproc(a)<ENTER>
     - (col>1)myproc(a)<ENTER>
     - (col>1)myproc(a)<ENTER>
     - case a: <ENTER>
     - default: <ENTER>
     -  if (...) {<ENTER>
     -  if (...) <ENTER>
     -  if (...) ++i; else <ENTER>
     -  if (...) ++i; else <ENTER>
     -  myproc (...) {<ENTER>
     -  statement;
         {<ENTER>
     -  if (a && b
     -  if (a && b,b
     -  <ENTER>  no code above
     -  int a,
     -  if {
           }<ENTER>
     -  {
        }<ENTER>
     - for (;<ENTER>;<ENTER>)
     - for (<ENTER>;;<ENTER>)
     - for (i=1;i<j;<ENTER>
     - if (a<b) {
          x=1;
       } else if( sfsdfd) {<ENTER>}

     {sdfsdf;
      ddd


*/

   _str enter_cmd=name_on_key(ENTER);
   if (enter_cmd=='nosplit-insert-line') {
      _end_line();
   }
   /*
       Handle a few special cases where line begins with
         close brace, "case", "default","public", "private",
         and "protected".
   */
   {
      save_pos(p2);
      //first_non_blank();
      begin_word();
      if (orig_col<=p_col) {
         cfg=_clex_find(0,'g');
         if (cfg!=CFG_COMMENT && cfg!=CFG_STRING) {
            word=cur_word(junk);
            ch=get_text();
            if (ch=="}") {
               right();
               col=c_endbrace_col(be_style);
               if (col) {
                  restore_pos(p);
                  return(col);
               }
            } else if (cfg==CFG_KEYWORD || (word=='signals' && ext_inherits_from('c'))) {
               if (word=='public' || word=='private' || word=='protected' ||
                   word=='signals') {
                  int class_col=find_class_col();
                  if (class_col) {
                     restore_pos(p);
                     return(class_col);
                  }
               } else if (word=='case' || word=='default') {
                  col=_c_last_switch_col();
                  if ( col) {
                     if ((indent_case && indent_case!='') || (be_style&VS_C_OPTIONS_STYLE2_FLAG)) {
                        col=col+syntax_indent;
                     }
                     restore_pos(p);
                     return(col);
                  }
               }

            }
         }
      }
      restore_pos(p2);
   }

   // Are we in an embedded context?
   // Then find the beginning of the embedded code
   long embedded_start_pos=0;
   if (p_EmbeddedLexerName!='') {
      save_pos(p2);
      if (!_clex_find(0,'-S')) {
         embedded_start_pos=_QROffset();
      }
      restore_pos(p2);
   }

   int nesting=0;
   int OpenParenCol=0;
   if (p_col==1) {
      up();_end_line();
   } else {
      left();
   }

   int status=search("[{;}:()]|with|if|elsif|elseif|while|lock|for|foreach|switch|using","@r-");
   for (;;) {
      if (status) {
         if (nesting<0) {
            restore_pos(p);
            return(OpenParenCol+1/*+def_c_space_after_paren*/);
         }
         return(NoSyntaxIndentCase(non_blank_col,orig_linenum,orig_col,p,syntax_indent));
      }

      if (_QROffset() < embedded_start_pos && p_embedded) {
         // we are embedded in HTML and hit script start tag
         //return(NoSyntaxIndentCase(non_blank_col,orig_linenum,orig_col,p,syntax_indent));
         restore_pos(p);
         if (nesting<0) {
            return(OpenParenCol+1/*+def_c_space_after_paren*/);
         }
         /* Cases:
              <%
                  <ENTER>
              %>
              <? <ENTER>
              ?>
              <script ...> <ENTER>
              </script>

             <cfif
                 IsDate(...) and<ENTER>
             </cfif>
         */
         _str orig_EmbeddedLexerName=p_EmbeddedLexerName;
         // Look for first non blank that is in this embedded language
         first_non_blank();
         int ilen=_text_colc();
         //_message_box('xgot here');
         for(;;) {
            if (p_col>ilen) {
               //_message_box('got here');
               return(orig_col);
            } else if (orig_EmbeddedLexerName==p_EmbeddedLexerName && get_text():!=' ') {
               //refresh();_message_box('break col='p_col' l='p_line);
               break;
            }
            ++p_col;
         }
         col=p_col;
         restore_pos(p);
         return(col);
      }

      cfg=_clex_find(0,'g');
      if (cfg==CFG_COMMENT || cfg==CFG_STRING) {
         status=repeat_search();
         continue;
      }

      ch=get_text();
      //messageNwait('ch='ch);
      switch (ch) {
      case '(':
         if (!nesting && !OpenParenCol) {
            save_pos(p3);
#if 1
            typeless ss1,ss2,ss3,ss4;
            save_search(ss1,ss2,ss3,ss4);
            col=p_col;
            ++p_col;
            status=_clex_skip_blanksNpp();


            if (!(UseContOnParameters==1) &&
                !status && (p_line<orig_linenum ||
                            (p_line==orig_linenum && p_col<orig_col)
                           )) {
               col=p_col-1;
            } else {
               /*
                  case: Use continuation indent instead of lining up on
                  open paren.

                  aButton.addActionListener(<Enter here. No args follow>
                      a,
                      b,
               */
               restore_pos(p3);
               goto_point(_nrseek()-1);
               //if (_clex_skip_blanks('-')) return(0);
               //word=cur_word(junk);
               c_prev_sym2();
               if (c_sym_gtk()=='>') {
                  parse_template_args();
               }
               if (c_sym_gtk()==TK_ID && !pos(' 'c_sym_gtkinfo()' ',' with for foreach if elsif elseif switch while lock using ')) {
                  restore_pos(p3);
                  first_non_blank();
                  col=p_col+p_SyntaxIndent-1;
               }
            }
            restore_search(ss1,ss2,ss3,ss4);
#else
            save_search(ss1,ss2,ss3,ss4);
            col=p_col;
            ++p_col;
            status=_clex_skip_blanksNpp();
            if (!status && (p_line<orig_linenum ||
                            (p_line==orig_linenum && p_col<=orig_col)
                           )) {
               col=p_col-1;
            }
            restore_search(ss1,ss2,ss3,ss4);
#endif
            OpenParenCol=col;
            restore_pos(p3);
         }
         --nesting;
         status=repeat_search();
         continue;
      case ')':
         ++nesting;
         status=repeat_search();
         continue;
      default:
         if (nesting<0) {
            //messageNwait("nesting case");
            restore_pos(p);
            return(OpenParenCol+1/*+def_c_space_after_paren*/);
         }
      }
      if (nesting ) {
         status=repeat_search();
         continue;
      }
      if (_in_c_preprocessing()) {
         begin_line();
         status=repeat_search();
         continue;
      }
      //messageNwait("c_indent_col2: ch="ch);
      switch (ch) {
      case '{':
         //messageNwait("case {");
         int openbrace_col=p_col;
         int statdelim_linenum=p_line;
         /*
            Could have
              for (;
                    ;) {<ENTER>

              myproc ( xxxx ) {<ENTER>

              myproc (xxx ) {
                 int i,<ENTER>

              {<ENTER>

              else {<ENTER>

              else
                 {<ENTER>

              class name : public name2 {<ENTER>

              if ( xxx ) {<ENTER>

              if ( xxx )
                 {<ENTER>

              if ( xxx )
              {<ENTER>

              int array[]={
                 a,
                 b,<ENTER>

         */
         save_pos(p2);

         if (_isVarInitList(true)) {
            restore_pos(p2);
            first_non_blank();
            col=p_col;
            restore_pos(p);
            return(col+p_SyntaxIndent);
#if 0
            restore_pos(p2);
            begin_stat_col=c_begin_stat_col(false /* No RestorePos */,
                                            true /* skip first begin statement marker */,
                                            true /* return first non-blank */
                                            );
            restore_pos(p2);
            // Now check if there are any characters between the
            // beginning of the previous statement and the original
            // cursor position
            col=HandlePartialStatement(statdelim_linenum,
                                       syntax_indent,0,
                                       orig_linenum,orig_col);
            if (col) {
               restore_pos(p);
               return(col);
            }
#endif
         }

         restore_pos(p2);

         if (p_col==1) {
            up();_end_line();
         } else {
            left();
         }
         _clex_skip_blanksNpp('-');
         status=1;
         if (get_text()==')') {
            status=_find_matching_paren(def_pmatch_max_diff);
            save_pos(p3);
         }
         if (!status) {
            status=1;
            if (p_col==1) {
               up();_end_line();
            } else {
               left();
            }
            _clex_skip_blanksNpp('-');
            if (_clex_find(0,'g')==CFG_KEYWORD) {
               kwd=cur_word(junk);
               status=(int) !pos(' 'kwd' ',' with if elsif elseif while lock switch for foreach catch using ');
               // IF this is the beginning of a "if/while/switch/for" block
               if (!status) {
                  first_non_blank();
                  int block_col=p_col;
                  // Now check if there are any characters between the
                  // beginning of the previous statement and the original
                  // cursor position
                  restore_pos(p2);


                  col=HandlePartialStatement(statdelim_linenum,
                                             syntax_indent,syntax_indent,
                                             orig_linenum,orig_col);
                  if (col) {
                     restore_pos(p);
                     return(col);
                  }

                  restore_pos(p);
                  return(block_col+syntax_indent);
               }
            } else if (ext_inherits_from('java')) {
               /*

                   // case 1:  just blanks after open paren.  Use continuation indent.
                   aButton.addActionListener(
                       a,
                       b,
                       new ActionListener() {
                           public void actionPerformed(ActionEvent e) {
                               createdButtonFired(buttonIndex);
                           }
                       },
                       b,
                       );
                  // case 2:  First argument is new constructor
                  aButton.addActionListener(new ActionListener() {
                          public void actionPerformed(ActionEvent e) {
                              createdButtonFired(buttonIndex);
                          }
                      },
                      b,
                      );

               */
               // Check if we have a new construct
               kwd=cur_word(col);
               p_col=_text_colc(col,'I');
               if (p_col>1) {
                  left();
                  if (_clex_skip_blanks('-')) return(0);
                  word=cur_word(col);
                  if (word=='new') {
                     p_col=_text_colc(col,'I');
                     col=p_col;
                     first_non_blank();
                     if (col!=p_col) {
                        p_col+=p_SyntaxIndent;
                     }
                     col=p_col+p_SyntaxIndent;
                     restore_pos(p);
                     return(col);
                  }
               }
            }

            // Now check if there are any characters between the
            // beginning of the previous statement and the original
            // cursor position
            restore_pos(p2);
            col=HandlePartialStatement(statdelim_linenum,
                                       syntax_indent,syntax_indent,
                                       orig_linenum,orig_col);
            if (col) {
               restore_pos(p);
               return(col);
            }

            //  This open brace is to a function or method or some
            //  very strange preprocessing.
            restore_pos(p2); // Restore cursor to open brace
            first_non_blank();
            if (p_col==openbrace_col) {
               begin_stat_col=openbrace_col;
            } else {
               restore_pos(p3); // Restore cursor to open paren
               begin_stat_col=c_begin_stat_col(false /* No RestorePos */,
                                               false /* Don't skip first begin statement marker */,
                                               false /* Don't return first non-blank */
                                               );
               if ((be_style & VS_C_OPTIONS_STYLE2_FLAG) && def_style3_indent_all_braces) {
                  begin_stat_col+=syntax_indent;
               }
            }

            if (begin_stat_col==1 && !indent_fl) {
               restore_pos(p);
               return(1);
            }
            restore_pos(p);
            if ((be_style & VS_C_OPTIONS_STYLE2_FLAG) && def_style3_indent_all_braces) {
               return(begin_stat_col);
            }
            return(begin_stat_col+syntax_indent);
         }
         restore_pos(p2);
         // Now check if there are any characters between the
         // beginning of the previous statement and the original
         // cursor position


         // Now check if there are any characters between the
         // beginning of the previous statement and the original
         // cursor position
         restore_pos(p2);
         col=HandlePartialStatement(statdelim_linenum,
                                    syntax_indent,syntax_indent,
                                    orig_linenum,orig_col);
         if (col) {
            restore_pos(p);
            return(col);
         }

         /*
             Probably have one of these case here

              {<ENTER>

              else {<ENTER>

              else
                 {<ENTER>

              class name : public name2 {<ENTER>

              if (a<b) x=1; else {<ENTER>}

              if (a<b) {
                 x=1;
              } else {<ENTER>}

         */
         restore_pos(p2);
         if (style2) {
            first_non_blank();
            // IF the open brace is the first character in the line
            if (openbrace_col==p_col) {

               begin_stat_col=c_begin_stat_col(false /* No RestorePos */,
                                               true /* skip first begin statement marker */,
                                               true /* return first non-blank */
                                               );
               // IF there is stuff between the previous statement and
               //    this statement, we must be in a class/struct
               //    definition.  IF/While/FOR Etc. cases have been
               //    handled above.
               if (openbrace_col!=p_col || statdelim_linenum!=p_line) {
                  restore_pos(p);
                  return(begin_stat_col+syntax_indent);
               }
               // We could check here for extra stuff after the
               // open brace
               restore_pos(p);
               return(openbrace_col);
            }
            restore_pos(p2);
         }
         begin_stat_col=c_begin_stat_col(false /* No RestorePos */,
                                         true /* skip first begin statement marker */,
                                         true /* return first non-blank */
                                         );
         restore_pos(p);
         return(begin_stat_col+syntax_indent);

      case ';':
         //messageNwait("case ;");
         save_pos(p2);
         statdelim_linenum=p_line;
         begin_stat_col=c_begin_stat_col(false /* RestorePos */,
                                    true /* skip first begin statement marker */,
                                    true /* return first non-blank */
                                    );
         /* IF there is extra stuff before the beginning of this
               statement
            Example
                x=1;y=2;<ENTER>
                       OR
                for (x=1;<ENTER>
            NOTE:  The following code fragrament does not work
                   properly.
                for (i=1;i<j;++i) ++i;<ENTER>
                for (i=1;
                     i<j;<ENTER>
         */
         word=cur_word(junk);
         if (word=='for') {
            // Here we try to indent after open brace for
            // loop unless the cursor is after the close paren.
            get_line_raw(line);line=expand_tabs(line);
            col=pos('(',line,1,p_rawpos);
            if (!col) {
               col=p_col;
               restore_pos(p);
               return(col+syntax_indent);
            }
            int result_col=col;
            p_col=col+1;
            search('[~ \t]','@r');
            cfg=_clex_find(0,'g');
            if (get_text()!='' && cfg!=CFG_COMMENT && cfg!=CFG_STRING) {
               ++result_col;
            }
            p_col=col;
            status=find_matching_paren();
            // IF cursor is after close paren of for loop
            if (!status && (orig_linenum>p_line ||
                            (p_line==orig_linenum && orig_col>p_col)
                           )
               ) {
               // Cursor is after close paren of for loop.
               //messageNwait('f1');
               restore_pos(p);
               return(begin_stat_col);
            }
            // Align cursor after open brace of for loop
            restore_pos(p);
            return(result_col+1);
         }
         restore_pos(p2);

         // Now check if there are any characters between the
         // beginning of the previous statement and the original
         // cursor position
         col=HandlePartialStatement(statdelim_linenum,
                                    syntax_indent,syntax_indent,
                                    orig_linenum,orig_col);
         if (col) {
            restore_pos(p);
            return(col);
         }
         restore_pos(p);
         return(begin_stat_col);
      case '}':
         //messageNwait("case }");
         /*
            Don't forget to test

            if (i<j)
               {
               }<ENTER>

            if (i<j)
               {
               }
            else
               {
               }<ENTER>


         */
         statdelim_linenum=p_line;
         save_pos(p2);
         /*
            Check if we are in a variable initialization list.
            We don't want to handle this with the HandlePartialStatement statement.
             MYRECORD array[]={
                {a,b,c}
                ,{a,b,c},
                b,<ENTER>
                <End UP HERE, ALIGNED WITH b>

         */
         right();
         _clex_skip_blanks();
         if (get_text()==',') {
            restore_pos(p2);
         } else {
            restore_pos(p2);
            /* Now check if there are any characters between the
               beginning of the previous statement and the original
               cursor position

               Could have
                 struct name {
                 } name1, <ENTER>

                 myproc() {
                 }
                    int i,<ENTER>
            */
            col=HandlePartialStatement(statdelim_linenum,
                                       syntax_indent,syntax_indent,
                                       orig_linenum,orig_col);
            if (col) {
               restore_pos(p);
               return(col);
            }
         }



         /*
             Handle the following cases
             for (;;)
                 {
                 }<ENTER>

                 {
                 }<ENTER>

             MYRECORD array[]={
                {a,b,c}<ENTER>

             MYRECORD array[]={
                {a,b,c}
                ,{a,b,c}<ENTER>

         */
         restore_pos(p2);
         ++p_col;
         boolean style3_MustBackIndent=false;
         col=c_endbrace_col2(be_style,style3_MustBackIndent);
         if (col) {
            if (!style2 || !style3_MustBackIndent) {
               restore_pos(p);
               return(col);
            }
            col-=syntax_indent;
            if (col<1) col=1;
            restore_pos(p);
            return(col);
         }
         restore_pos(p2);
         if (!style2 || !style3_MustBackIndent) {
            col=p_col;
            restore_pos(p);
            return(col);
         }
         col=p_col-syntax_indent;
         if (col<1) col=1;
         restore_pos(p);
         return(col);
      case ':':
         //messageNwait("case :");
         if (ext_inherits_from('e')){
            // Watch out for :==,:!=, :+, :<=, :>=
            ch=get_text(1,(int)point('s')+1);
            if(ch=='=' || ch=='!' || ch=='<' || ch=='>' || ch=='+' ||  '[' /* :[ ]  Slick-C operator*/) {
               status=repeat_search();
               continue;
            }
         }
         if (p_col!=1) {
            left();
            if (get_text()==":") {
               // skip ::
               //messageNwait('skip ::');
               status=repeat_search();
               continue;
            }
            right();
         }

         save_pos(p2);
         typeless t1,t2,t3,t4;
         save_search(t1,t2,t3,t4);
         boolean bool=_isQmarkExpression();
         //messageNwait('isQmark='bool);
         restore_pos(p2);
         restore_search(t1,t2,t3,t4);
         if (bool) {
            //skip this question mark expression colon
            /*
               NOTE: We could handle the following case better here:
               myproc(b,
                     (c)?s:<ENTER>
                     )
               which is different from
               myproc(b,
                     (c)?s:t,<ENTER>
                     )
            */
            status=repeat_search();
            continue;
         }

         /* Now check if there are any characters between the
            beginning of the previous statement and the original
            cursor position

            Could have
             case 'a':
                 int i,<ENTER>

            MyConstructor(): a(1),<ENTER>b(2)
         */
         col=HandlePartialStatement(statdelim_linenum,
                                    syntax_indent,syntax_indent,
                                    orig_linenum,orig_col);
         if (col) {
            //messageNwait('c1');
            restore_pos(p);
            return(col);
         }
         //messageNwait('c2');

         restore_pos(p2);


         /*

             default:<ENTER>
             case ???:<ENTER>
             (abc)? a: b;<ENTER>
             class name1:public<ENTER>
         */
         begin_stat_col=c_begin_stat_col(false /* RestorePos */,
                                    true /* skip first begin statement marker */,
                                    true /* return first non-blank */,
                                    1
                                    );

         if (p_line==orig_linenum) {
            word=cur_word(junk);
            if (word=='case' || word=='default') {
               first_non_blank();
               // IF the 'case' word is the first non-blank on this line
               if (p_col==begin_stat_col) {
                  col=p_col;
                  restore_pos(p);
                  //messageNwait('c3');
                  return(col);
               }
            }
         }
         //messageNwait('c4');
         restore_pos(p);
         return(begin_stat_col+syntax_indent);
      default:
         if (cfg==CFG_KEYWORD) {
            /*
               Cases
                 if ()
                    if () <ENTER>
                 for <ENTER>

            */
            first_non_blank();
            col=p_col+syntax_indent;
            restore_pos(p);
            return(col);
         }
      }

      status=repeat_search();
   }

}

int _c_get_syntax_completions(var words)
{
   typeless space_words;
   if (ext_inherits_from('phpscript')) {
      space_words = php_space_words;
   } else if (ext_inherits_from('idl')) {
      space_words = idl_space_words;
   } else if (ext_inherits_from('cs')) {
      space_words = cs_space_words;
   } else if (ext_inherits_from('js') || ext_inherits_from('cfscript')) {
      space_words = javascript_space_words;
   } else if (ext_inherits_from('java')) {
      space_words = java_space_words;
   } else {
      space_words = cpp_space_words;
   }

   return AutoCompleteGetSyntaxSpaceWords(words,space_words,0);
}
boolean c_else_followed_by_brace_else(_str word)
{
   // this must be an else if
   if (!pos('else',word) || !pos('if',word)) {
      return false;
   }

   int status=0;
   boolean found_brace_else=false;
   typeless p;
   typeless s1,s2,s3,s4;
   save_pos(p);
   save_search(s1,s2,s3,s4);

   first_non_blank();

   status=search('[^ \t\n\r]','@-rXc');
   if (status || get_text() != '}') {
      restore_search(s1,s2,s3,s4);
      restore_pos(p);
      return false;
   }

   int close_brace_col = p_col;

   restore_pos(p);
   _end_line();
   status=search('[^ \t\n\r]','@rXc');
   if (status==0 && get_text()=='}' && p_col == close_brace_col) {
      right();
      c_next_sym();
      if (c_sym_gtkinfo()=='else') {
         found_brace_else=true;
      }
   }

   restore_search(s1,s2,s3,s4);
   restore_pos(p);
   return found_brace_else;
}
static int c_expand_space(int syntax_indent,int be_style,int indent_fl)
{
   int status=0;
   _str orig_line='';
   get_line(orig_line);
   _str line=strip(orig_line,'T');
   _str orig_word=strip(line);
   if ( p_col!=text_col(_rawText(line))+1 ) {
      return(1);
   }
   boolean if_special_case=false;
   boolean else_special_case=false;
   boolean pick_else_or_else_if=false;
   _str brace_before='';
   _str aliasfilename='';
   int java=0, idl=0, php=0,cs=0;
   if (ext_inherits_from('java') || ext_inherits_from('cs')) {
      java=1;
   } else if (ext_inherits_from('js') || ext_inherits_from('cfscript') || ext_inherits_from('as')) {
      java=2;
   }
   _str word='';
   if (ext_inherits_from('phpscript')) {
      word=min_abbrev2(orig_word,php_space_words,name_info(p_index),aliasfilename);
      php=1;
   } else if (ext_inherits_from('idl')) {
      idl=1;
      word=min_abbrev2(orig_word,idl_space_words,name_info(p_index),aliasfilename);
   } else if (ext_inherits_from('cs')) {
      word=min_abbrev2(orig_word,cs_space_words,name_info(p_index),aliasfilename);
      cs=1;
   } else if (java==2) {
      word=min_abbrev2(orig_word,javascript_space_words,name_info(p_index),aliasfilename);
   } else if (java) {
      word=min_abbrev2(orig_word,java_space_words,name_info(p_index),aliasfilename);
   } else {
      word=min_abbrev2(orig_word,cpp_space_words,name_info(p_index),aliasfilename);
   }
   if (aliasfilename==null) return(0);
   if (aliasfilename!='' && word!='') {
      typeless mult_line_info;
      if (orig_word:==word && orig_word==get_alias(word,mult_line_info,1,aliasfilename)) {
         _insert_text(' ');
         return(0);
      }
      _str line_prefix='';
      int col=p_col-_rawLength(orig_word);
      if (col==1) {
         line_prefix='';
      }else{
         line_prefix=indent_string(col-1);
      }
      replace_line(line_prefix);
      p_col=col;
      return(expand_alias(word,'',aliasfilename));
   }
   if ( word=='') {
      // Check for } else
      _str first_word, second_word, rest;
      parse orig_line with first_word second_word rest;
      if (!def_always_prompt_for_else_if && first_word=='}' && second_word!='' && rest=='' && second_word=='else') {
         //Can't force user to use modal dialog insead of just typing "} else {"
         //We need a modeless dialog so user can keep typing.
         return(1);
      } else if (!def_always_prompt_for_else_if && second_word=='' && length(first_word)>1 && first_word:=='}else') {
         //Can't force user to use modal dialog insead of just typing "}else {"
         //We need a modeless dialog so user can keep typing.
         return(1);
      } else if (first_word=='}' && second_word!='' && rest=='' && second_word:==substr('else',1,length(second_word))) {
         brace_before='} ';
         first_word=second_word;
         pick_else_or_else_if=true;
      } else if (second_word=='' && length(first_word)>1 && first_word:==substr('}else',1,length(first_word))) {
         brace_before='}';
         first_word=substr(first_word,2);
         pick_else_or_else_if=true;
      } else if (php && first_word=='}' && second_word!='' && rest=='' && second_word:==substr('elseif',1,length(second_word))) {
         word='} elseif';
         if_special_case=true;
      } else if (php && second_word=='' && length(first_word)>1 && first_word:==substr('}elseif',1,length(first_word))) {
         word='}elseif';
         if_special_case=true;
      } else if (!php && !idl && first_word=='}' && second_word!='' && rest=='' && second_word:==substr('finally',1,length(second_word))) {
         word='} finally';
         else_special_case=true;
      } else if (!php && !idl && second_word=='' && length(first_word)>1 && first_word:==substr('}finally',1,length(first_word))) {
         word='}finally';
         else_special_case=true;
      // Check for else if or } else if
      } else if (first_word=='else' && orig_word==substr('else if',1,length(orig_word))) {
         word='else if';
         if_special_case=true;
      } else if (second_word=='else' && rest!='' && orig_word==substr('} else if',1,length(orig_word))) {
         word='} else if';
         if_special_case=true;
      } else if (first_word=='}else' && second_word!='' && orig_word==substr('}else if',1,length(orig_word))) {
         word='}else if';
         if_special_case=true;
      } else {
         return(1);
      }
   } else if (!def_always_prompt_for_else_if && orig_word=='else' && word=='else') {
      //Can't force user to use modal dialog insead of just typing "}else {"
      //We need a modeless dialog so user can keep typing.
      return(1);
   } else if (orig_word=='else' && word=='else') {
      pick_else_or_else_if=true;
   }
   if (pick_else_or_else_if) {
      if (php || ext_inherits_from('rul')) {
         word=min_abbrev2('els',php_space_words,name_info(p_index),'');
      } else {
         word=min_abbrev2('els',else_space_words,name_info(p_index),'');
      }
      switch (word) {
      case 'else':
         word=brace_before:+word;
         else_special_case=true;
         break;
      case 'elseif':
      case 'else if':
         word=brace_before:+word;
         if_special_case=true;
         break;
      default:
         return(1);
      }
   }
   clear_hotspots();
   _str maybespace=(be_style & VS_C_OPTIONS_NO_SPACE_BEFORE_PAREN)?'':' ';
   line=substr(line,1,length(line)-length(orig_word)):+word;
   int width=text_col(_rawText(line),_rawLength(line)-_rawLength(word)+1,'i')-1;
   int style1=be_style & VS_C_OPTIONS_STYLE1_FLAG;
   int style2=be_style & VS_C_OPTIONS_STYLE2_FLAG;
   _str e1=' {';
   if (! ((word=='do' || word=='try' || word=='finally') && !style2 && !style1) ) {
      if ( (be_style & (VS_C_OPTIONS_STYLE1_FLAG|VS_C_OPTIONS_STYLE2_FLAG)) ||
         ! (be_style & VS_C_OPTIONS_BRACE_INSERT_FLAG) ) {
         e1='';
      } else if (word=='}else' || word=='}finally') {
         e1='{';
      }
   }
   if ( word=='main' ) {
      if (ext_inherits_from('java')) {
         typeless p;
         save_pos(p);
         int col=find_class_col();
         restore_pos(p);
         // If there is no class in this file
         if (!col) {
            replace_line("public class "strip_filename(p_buf_name,"pe")" {");
            width=syntax_indent;
            insert_line(indent_string(width)'public static void main (String args[]) {');
         } else {
            replace_line(indent_string(width)'public static void main (String args[]) {');
         }
         insert_line('');
         insert_line(indent_string(width)'}');
         if (!col) {
            insert_line("}");
            up();
         }
         up();p_col=width+((indent_fl)?syntax_indent:0)+1;
         return(0);
      } else if (ext_inherits_from('cs')) {
         typeless p;
         save_pos(p);
         int col=find_class_col();
         restore_pos(p);
         // If there is no class in this file
         if (!col) {
            replace_line("using System;");
            insert_line("class "strip_filename(p_buf_name,"pe")" {");
            width=syntax_indent;
            insert_line(indent_string(width)'public static void Main (string []args) {');
         } else {
            replace_line(indent_string(width)'public static void Main (string []args) {');
         }
         insert_line('');
         insert_line(indent_string(width)'}');
         if (!col) {
            insert_line("}");
            up();
         }
         up();p_col=width+((indent_fl)?syntax_indent:0)+1;
         return(0);
      } else {
         status=c_insert_main();
      }
   } else if ( word=='if' || word=='elseif' || word=='else if' || if_special_case) {
      replace_line(line:+maybespace:+'()'e1);
      maybe_insert_braces(syntax_indent,be_style,width,word,c_else_followed_by_brace_else(word));
      add_hotspot();
   } else if ( word=='else') {
      typeless p;
      typeless s1,s2,s3,s4;
      save_pos(p);
      save_search(s1,s2,s3,s4);
      up();_end_line();
      search('[^ \t\n\r]','@-rXc');
      if (get_text()=='}') {
         be_style |= VS_C_OPTIONS_BRACE_INSERT_FLAG;
      } else {
         e1=' ';
         be_style &= ~VS_C_OPTIONS_BRACE_INSERT_FLAG;
      }
      restore_search(s1,s2,s3,s4);
      restore_pos(p);
      replace_line(line:+e1);
      maybe_insert_braces(syntax_indent,be_style,width,word);
      _end_line();
   } else if ( word=='finally' || else_special_case) {
      replace_line(line:+e1);
      maybe_insert_braces(syntax_indent,be_style|VS_C_OPTIONS_BRACE_INSERT_FLAG,width,word);
      _end_line();
   } else if ( word=='for' || (word=='with' && java==2)) {
      replace_line(line:+maybespace'()'e1);
      maybe_insert_braces(syntax_indent,be_style,width,word);
      add_hotspot();
   } else if ( word=='while' ) {
      replace_line(line:+maybespace'()'e1);
      maybe_insert_braces(syntax_indent,be_style,width,word);
      add_hotspot();
   } else if ( word=='using' ) {
      if (_in_function_scope()) {
         replace_line(line:+maybespace'()'e1);
         maybe_insert_braces(syntax_indent,be_style,width,word);
         add_hotspot();
      } else {
         replace_line(line' ');
         _end_line();
      }
   } else if ( word=='foreach' ) {
      replace_line(line:+maybespace'( in )'e1);
      maybe_insert_braces(syntax_indent,be_style,width,word);
      add_hotspot();
      p_col+=4;
      add_hotspot();
      p_col-=4;
   } else if ( word=='lock' || word=='fixed' ) {
      replace_line(line:+maybespace'()'e1);
      maybe_insert_braces(syntax_indent,be_style,width,word);
      add_hotspot();
   } else if ((idl || php) && pos(' 'word' ',' exception interface struct class module ' )) {
      replace_line(line' 'e1);
      maybe_insert_braces(syntax_indent,be_style,width,word);
      if (maybespace:==' ') {
         left();
      }
   } else if (idl && word=='union') {
      replace_line(line'  switch'maybespace'()'e1);
      maybe_insert_braces(syntax_indent,be_style,width,word);
      left();
   } else if (word=='enum' && ext_inherits_from('java')) {
      replace_line(line:+maybespace:+e1);
      maybe_insert_braces(syntax_indent,be_style,width,word);
      left();
      add_hotspot();
   } else if (word=='@interface') {
      replace_line(line:+maybespace:+e1);
      maybe_insert_braces(syntax_indent,be_style,width,word);
      left();
      add_hotspot();
   } else if (idl && word=='sequence') {
      replace_line(line:+maybespace:+'<>');
      _end_line();left();
   } else if ( (word=='public' || word=='private' || word=='protected') && !java && p_extension!='tagdoc') {
      replace_line(line':');_end_line();
      _c_do_colon();
   } else if ( word=='switch' ) {
      replace_line(line:+maybespace'()'e1);
      maybe_insert_braces(syntax_indent,be_style,width,word);
      add_hotspot();
   } else if ( word=='do' ) {
      // Always insert braces for do loop unless braces are on separate
      // line from do and while statements
      replace_line(line:+e1);
      if ( ! style2 ) {
         if (style1 ) {
            insert_line(indent_string(width)'{');
         }
         insert_line(indent_string(width)'} while':+maybespace'(  );');
         _end_line();
         p_col -= 3;
         add_hotspot();
         up();
      } else if ( style2 ) {
         if (be_style & VS_C_OPTIONS_BRACE_INSERT_FLAG) {
            insert_line(indent_string(width+syntax_indent)'{');
            insert_line(indent_string(width+syntax_indent)'}');
            insert_line(indent_string(width)'while':+maybespace'(  );');
            _end_line();
            p_col -= 3;
            add_hotspot();
            up(2);
            syntax_indent=0;
         } else {
            insert_line(indent_string(width)'while'maybespace:+'(  );');
            _end_line();
            p_col -= 3;
            add_hotspot();
            up(1);
            //syntax_indent=0
         }
      }
      nosplit_insert_line();
      p_col=p_col+syntax_indent;
      add_hotspot();
   } else if ( word=='try' ) {
      replace_line(line:+e1);
      if ( ! style2 ) {
         if (style1 ) {
            insert_line(indent_string(width)'{');
         }
         insert_line(indent_string(width)'} catch':+maybespace'(  )'e1);
         _end_line();
         p_col-=length(e1)+2;
         add_hotspot();
         maybe_insert_braces(syntax_indent,be_style|VS_C_OPTIONS_BRACE_INSERT_FLAG,width,word);
         up(1);
      } else if ( style2 ) {
         insert_line(indent_string(width+syntax_indent)'{');
         insert_line(indent_string(width+syntax_indent)'}');
         insert_line(indent_string(width)'catch':+maybespace'(  )'e1);
         _end_line();
         p_col -= (length(e1)+2);
         add_hotspot();
         maybe_insert_braces(syntax_indent,be_style|VS_C_OPTIONS_BRACE_INSERT_FLAG,width,word);
         up(2);
         syntax_indent=0;
      }
      nosplit_insert_line();
      p_col=p_col+syntax_indent;
      add_hotspot();
   } else if ( word=='printf' ) {
      replace_line(indent_string(width)'printf("');
      _end_line();
   } else if ( java==2 && (word=='export' || word=='function' || word=='import' || word=='var')) {
      replace_line(indent_string(width)word' ');_end_line();
   } else if ( word=='return' ) {
      boolean IsVoid=false;
      _UpdateContext(true);
      int context_id=tag_current_context();
      if (context_id>0) {
         _str tag_type='';
         tag_get_detail2(VS_TAGDETAIL_context_type,context_id,tag_type);
         if (tag_tree_type_is_func(tag_type)) {
            _str return_type='';
            tag_get_detail2(VS_TAGDETAIL_context_return,context_id,return_type);
            if (return_type=='void') {
               IsVoid=true;
            }
         }
      }
      if (orig_word=='return') {
         if (IsVoid) {
            keyin(';');
         }else{
            keyin(' ');
         }
      } else {
         replace_line(indent_string(width)'return');
         _end_line();
         if (IsVoid) {
            keyin(';');
         } else {
            keyin(' ');
         }
      }
   } else if ( word=='continue' || word=='break' ) {
      // Java allows labels to follow continue or break
      if (java) {
         if (orig_word==word) {
            replace_line(indent_string(width)word' ');
         } else {
            replace_line(indent_string(width)word);
         }
      } else {
         replace_line(indent_string(width)word';');
      }
      _end_line();
   } else if ( word=='case' ) {
      if ( name_on_key(ENTER):=='nosplit-insert-line' ) {
         replace_line(indent_string(width)word' :');
         _end_line();_c_do_colon();p_col=p_col-1;
         if ( ! _insert_state() ) _insert_toggle();
      } else {
#if 1
         // Code which inserts case
         replace_line(indent_string(width)word' :');
         _end_line();_c_do_colon();_rubout();
#else
         // Code which inserts case and colon and
         // puts user in insert mode.
         replace_line(indent_string(width)word' :');
         _end_line();_c_do_colon();p_col=p_col-1;
         if ( ! _insert_state() ) _insert_toggle();
#endif
      }
   } else if ( word=='default' ) {
      replace_line(indent_string(width)word':');_end_line();
      _c_do_colon();
   } else if ( pos(' 'word' ',EXPAND_WORDS) ) {
      replace_line(indent_string(width)word' ');
      _end_line();
   } else if (java==1 && pos(' 'word' ',JAVA_ONLY_EXPAND_WORDS) ) {
      replace_line(indent_string(width)word' ');
      _end_line();
   } else if (cs==1 && pos(' 'word' ',CS_ONLY_EXPAND_WORDS) ) {
      replace_line(indent_string(width)word' ');
      _end_line();
   } else if (idl==1 && pos(' 'word' ',IDL_ONLY_EXPAND_WORDS) ) {
      replace_line(indent_string(width)word' ');
      _end_line();
   } else if (php==1 && pos(' 'word' ',PHP_ONLY_EXPAND_WORDS) ) {
      replace_line(indent_string(width)word' ');
      _end_line();
   } else {
     status=1;
   }
   return(status);
}

static int c_expand_begin(int expand,int syntax_indent,int be_style,int indent_fl)
{
   int brace_indent=0;
   keyin('{');
   _str line;
   get_line(line);
   int pcol=_text_colc(p_col,'P');
   _str last_word='';
   typeless AfterKeyinPos;
   save_pos(AfterKeyinPos);

   // check that brace expansion is enabled
   if (!def_expand_brace_syntax) {
      expand = 0;
   }

   // first, back up and look for a parenthesized expression
   // which would be part of the if, while, or for statement
   left();
   left();
   _clex_skip_blanks('-');
   if (get_text()!=')' || find_matching_paren() != 0) {
      restore_pos(AfterKeyinPos);
   }

   // compute the simple indentation column for this line
   first_non_blank();
   int indent_col = p_col;

   // now attempt to find a brace matching the brace we just put in
   // if it falls in a column that matches the expected indentation
   // then do not insert the closing brace
   restore_pos(AfterKeyinPos);
   if (expand && !find_matching_paren()) {
      if (p_col >= indent_col && p_col <= indent_col+p_SyntaxIndent) {
         expand=0;
      }
   }
   restore_pos(AfterKeyinPos);

   /*
        Don't insert end brace for these cases in a variable initializer
        object array={
           {<DONT EXPAND THIS>
        }
        object array={
           a,{<DONT EXPAND THIS>
        }

   */
   left();
   if (p_col==1) {
      up();_end_line();
   } else {
      left();
   }
   _clex_skip_blanksNpp('-');
   if (get_text()==',') {
      restore_pos(AfterKeyinPos);
      return(0);
   }
   if (get_text()=='{') {
      // This won't work for C because of function variable declarations but should work pretty well for C++
      // Worst case, user has to type close brace
      if (p_col==1) {
         up();_end_line();
      } else {
         left();
      }
      _clex_skip_blanksNpp('-');
      if (get_text()!=')') {
         restore_pos(AfterKeyinPos);
         return(0);
      }
   }
   restore_pos(AfterKeyinPos);

   int old_linenum=p_line;
   int col=0, old_col=p_col;
   int begin_brace_col=0;
   int status=_clex_skip_blanks();
   boolean end_brace_is_last_char=status || p_line>old_linenum;
   restore_pos(AfterKeyinPos);

   if ( line!='{' ) {
      if (!end_brace_is_last_char) {
         return(0);
      }
   } else if ( ! (be_style & VS_C_OPTIONS_STYLE2_FLAG) ) {
      /*
          Now that "class name<ENTER>" usually indents, we need
          the begin brace to be moved correctly to align under the
          "class" keyword.
      */
      typeless p;
      save_pos(p);
      left();
      //begin_brace_col=p_col;
      col= find_block_col();
      if (!col) {
         restore_pos(p);left();
         col=c_begin_stat_col(true,true,true);
      } else {
         // Indenting for class/struct/interface/variable initialization
         /*style=(be_style & VS_C_OPTIONS_STYLE2_FLAG);
         if (style!=0) {
            col=begin_brace_col;
         }*/
      }
      restore_pos(p);
      if (col) {
         replace_line(indent_string(col-1)'{');
         _end_line();save_pos(AfterKeyinPos);
      }

   } else if ( (be_style & VS_C_OPTIONS_STYLE2_FLAG) ) {
      /*
         A few customers like the way 1.7 let them type braces
         for functions indented.

         Brief does not do this.

      */
      /*
          Now that "class name<ENTER>" usually indents, we need
          the begin brace to be moved correctly to align under the
          "class" keyword.
      */
      typeless p;
      save_pos(p);
      left();
      begin_brace_col=p_col;
      col= find_block_col();
      if (!col) {
         restore_pos(p);left();
         col=c_begin_stat_col(true,true,true);
         if ((be_style & VS_C_OPTIONS_STYLE2_FLAG) && def_style3_indent_all_braces) {
            col+=syntax_indent;
         }
      } else {
         // Indenting for class/struct/interface/variable initialization
         int style=(be_style & (VS_C_OPTIONS_STYLE1_FLAG|VS_C_OPTIONS_STYLE2_FLAG));
         if (style!=0) {
            col=begin_brace_col;
         }
      }
      restore_pos(p);
      if (col) {
         replace_line(indent_string(col-1)'{');
         _end_line();save_pos(AfterKeyinPos);
      }

   }
   first_non_blank();
   if ( expand ) {
      col=p_col-1;
      if ( (col && (be_style & VS_C_OPTIONS_STYLE2_FLAG)) || (! (indent_fl+col)) ) {
         syntax_indent=0;
      }
      insert_line(indent_string(col+brace_indent));
      c_endbrace();
      restore_pos(AfterKeyinPos);//_end_line();
      if ((be_style & VS_C_OPTIONS_BRACE_INSERT_LINE_FLAG) ) {
         _end_line();
         c_enter();
      }
#if 0
      if ( be_style & VS_C_OPTIONS_BRACE_INSERT_LINE_FLAG ) {
         insert_line indent_string(col+syntax_indent)
      }
      insert_line indent_string(col+brace_indent)'}'
      up;_end_line
#endif
   } else {
      restore_pos(AfterKeyinPos);//_end_line();
   }
   typeless done_pos;
   save_pos(done_pos);
   if (ext_inherits_from('c') || ext_inherits_from('java') || ext_inherits_from('cs')) {
      restore_pos(AfterKeyinPos);
      _str class_name='';
      _str implement_list='';
      _str class_type_name;
      int vsImplementFlags=0;
      indent_col=c_parse_class_definition(class_name,class_type_name,implement_list,vsImplementFlags,AfterKeyinPos);
      if (!indent_col) {
         restore_pos(done_pos);
         return(0);
      }
      restore_pos(AfterKeyinPos);
      /*
         For simplicity, remove blank line that was inserted
      */
      if (expand && (be_style & VS_C_OPTIONS_BRACE_INSERT_LINE_FLAG) ) {
         down();
         _delete_line();
         restore_pos(AfterKeyinPos);
      }
      int count;
      //messageNwait('class_name='class_name' implement_list='implement_list);
      int context_id=tag_current_context();
      _str outer_class="";
      _str tag_name="";
      if (context_id>0) {
         tag_get_detail2(VS_TAGDETAIL_context_class, context_id, outer_class);
      }
      if (class_name=='') {
         class_name=implement_list;
      }
      count=_do_default_get_implement_list(class_name, outer_class,vsImplementFlags,false);
      _str lastext="";
      int index=0;
      boolean CursorDone=false;
      int c_access_flags = VS_TAGFLAG_private;
      int match_id;
      for (match_id=1;match_id<=count;++match_id) {
         VS_TAG_BROWSE_INFO cm;
         tag_get_match_info(match_id, cm);

         // Can't we get source comments?
         _str header_list[];header_list._makeempty();
         _ExtractTagComments(header_list,2000,cm.member_name,cm.file_name,cm.line_no,
                             cm.type_name, cm.class_name, indent_col);
         // generate the match signature for this function, not a prototype
         int akpos=_c_generate_function(cm,c_access_flags,header_list,null,
                                               indent_col,brace_indent,false);
         if (!CursorDone) {
            CursorDone=true;
            AfterKeyinPos = akpos;
         }
      }
      down();
      _str cur_line;
      get_line(cur_line);
      if (ext_inherits_from('c') && cur_line=='}') {
         replace_line(strip(cur_line,'T'):+';');
      }
      restore_pos(AfterKeyinPos);
   }
   return(0);
}
static _str prev_stat_has_semi()
{
   int status=1;
   up();
   if ( ! rc ) {
      int col=p_col;
      _end_line();
      _str line;
      get_line_raw(line);
      parse line with line '\#|/\*|//',(p_rawpos'r');
      /* parse line with line '{' +0 last_word ; */
      /* parse line with first_word rest ; */
      /* status=stat_has_semi() or line='}' or line='' or last_word='{' or first_word='case' */
      line=strip(line,'T');
      if (raw_last_char(line)==')') {
         typeless p;
         save_pos(p);
         p_col=text_col(line);
         status=_find_matching_paren(def_pmatch_max_diff);
         if (!status) {
            status=search('[~( \t]','@-r');
            if (!status) {
               if (!_clex_find(0,'g')==CFG_KEYWORD) {
                  status=1;
               } else {
                  _str junk,kwd=cur_word(junk);
                  status=(int) !pos(' 'kwd' ',' if do while switch for ');
               }
            }
         }
         restore_pos(p);
      } else {
         status=(int) (raw_last_char(line)!=')' && (int) !pos('(\}|)else$',line,1,p_rawpos'r'));
      }
      down();
      p_col=col;
   }
   return(status);
}
static _str stat_has_semi()
{
   _str line;
   get_line_raw(line);
   parse line with line '/*',p_rawpos;
   parse line with line '/\*|//',(p_rawpos'r');
   line=strip(line,'T');
   _str name=name_on_key(ENTER);
   return((raw_last_char(line):==';' || raw_last_char(line):=='}') &&
            (
               ! ((_will_split_insert_line()
                    ) && (p_col<=text_col(line) && arg(1)=='')
                   )
            )
         );

}
static void maybe_insert_braces(int syntax_indent,int be_style,int width,
                                _str word,boolean no_close_brace=false)
{
   int col=width+length(word)+3;
   if (be_style & VS_C_OPTIONS_NO_SPACE_BEFORE_PAREN) --col;
   if ( be_style & VS_C_OPTIONS_STYLE2_FLAG ) {
      width=width+syntax_indent;
   }
   if ( be_style & VS_C_OPTIONS_BRACE_INSERT_FLAG ) {
      int up_count=1;
      if ( be_style & (VS_C_OPTIONS_STYLE1_FLAG|VS_C_OPTIONS_STYLE2_FLAG) ) {
         up_count=up_count+1;
         insert_line(indent_string(width)'{');
      }
      if ( be_style & VS_C_OPTIONS_BRACE_INSERT_LINE_FLAG ) {
         up_count=up_count+1;
         if ( (be_style & (VS_C_OPTIONS_STYLE1_FLAG|VS_C_OPTIONS_STYLE2_FLAG))==VS_C_OPTIONS_STYLE2_FLAG) {
            insert_line(indent_string(width));
         } else {
            insert_line(indent_string(width+syntax_indent));
         }
      }
      _end_line();
      add_hotspot();
      if (no_close_brace) {
         up_count--;
      } else {
         insert_line(indent_string(width)'}');
      }
      up(up_count);
   }
   p_col=col;
   if ( ! _insert_state() ) _insert_toggle();
}
/*
   It is no longer necessary to modify this function to
   create your own main style.  Just define an extension
   specific alias.  See comment at the top of this file.

   NOTE: This function is not called for java.
*/
static _str c_insert_main()
{
   typeless expand, be_style, indent_fl, main_style;
   parse name_info(p_index) with . . . . be_style indent_fl main_style .;
   int syntax_indent=p_SyntaxIndent;
   if ( ! indent_fl ) {
      syntax_indent=0;
   }
   if ( main_style==2 ) {
      return(1);
   }
   _str maybespace=(be_style & VS_C_OPTIONS_NO_SPACE_BEFORE_PAREN)?'':' ';
   _str char_ptr = "char *";
   if (be_style & VS_C_OPTIONS_SPACE_AFTER_POINTER) {
      char_ptr = "char* ";
   }
   if (be_style & VS_C_OPTIONS_SPACE_SURROUNDS_POINTER) {
      char_ptr = "char * ";
   }

   if ( ! main_style /* K&R */) {
      replace_line('main':+maybespace'(argc, argv)');
      insert_line(indent_string(syntax_indent):+'int argc;');
      insert_line(indent_string(syntax_indent):+char_ptr:+'argv[];');
   } else {          /* ANSI / C++ */
#if __UNIX__
      // GNU c++ wants int return type
      replace_line('int main':+maybespace'(int argc, ':+char_ptr:+'argv[])');
#else
      replace_line('void main':+maybespace'(int argc, ':+char_ptr:+'argv[])');
#endif
   }
   insert_line('{');
   insert_line('');
   insert_line('}');
   up();p_col=syntax_indent+1;
   return(0);
}



defeventtab c_keys;
def '('=auto_functionhelp_key;
def '<'=auto_functionhelp_key;
def '.'=auto_codehelp_key;
def '>'=auto_codehelp_key;
def '='=auto_codehelp_key;
def '#'=c_pound;
def '@'=c_atsign;
def ':'=c_colon;
def '*'=c_asterisk;
def '/'=c_slash;
//def 'a'-'z'=c_identifier_key;
//def 'A'-'Z'=c_identifier_key;
//def '0'-'9'=c_identifier_key;
//def '_'=c_identifier_key;
//def del=c_identifier_key;
//def backspace=c_identifier_key;

